import os
import time

import numpy as np
import pandas as pd
import scanpy as sc
import scipy.sparse as ssp
from tqdm import tqdm

from cospar.tmap import _utils as tmap_util

from .. import help_functions as hf
from .. import logging as logg
from .. import settings
from .. import tool as tl
from .optimal_transport import optimal_transport_duality_gap, transport_stablev2


def refine_Tmap_through_cospar(
    MultiTime_cell_id_array_t1,
    MultiTime_cell_id_array_t2,
    proportion,
    transition_map,
    X_clone,
    initial_similarity,
    final_similarity,
    sparsity_threshold=0.1,
    normalization_mode=1,
):
    """
    This performs one iteration of coherent sparsity optimization.

    This is our core algorithm for Tmap inference. It updates a map
    by considering clones spanning multiple time points.

    Parameters
    ----------
    MultiTime_cell_id_array_t1: `np.array`
        An array of cell id sub_array, where each sub_array consists of
        clonally-related cell id's at different time points
    MultiTime_cell_id_array_t2: `np.array`
        An corresponding array of sub_array, where each sub_array are id's of
        cells that are clonally related to the corresponding sub_array at
        MultiTime_cell_id_array_t1.
    proportion: `list`
        A weight factor for each time point.
    transition_map: `np.array` or `sp.spmatrix`
        initialized transition map, or map from a previous iteration.
    X_clone: `sp.spmatrix`
        clonal matrix
    initial_similarity: `np.array`
        similarity matrix for all cells belonging
        to MultiTime_cell_id_array_t1
    final_similarity: `np.array`
        similarity matrix for all cells belonging
        to MultiTime_cell_id_array_t2
    sparsity_threshold: `float`, optional (default: 0.1)
        The relative threshold to remove noises in the updated transition map,
        in the range [0,1].
    normalization_mode: `int`, optional (default: 1)
        Normalization method. Choice: [0,1].
        0, single-cell normalization; 1, Clone normalization. The clonal
        normalization suppresses the contribution of large
        clones, and is much more robust.

    Returns
    -------
    smoothed_new_transition_map: `np.array`
    un_SM_transition_map: `np.array`
    """

    resol = 10 ** (-10)
    transition_map = hf.matrix_row_or_column_thresholding(
        transition_map, sparsity_threshold, row_threshold=True
    )

    if normalization_mode == 0:
        logg.hint("Single-cell normalization")
    if normalization_mode == 1:
        logg.hint("Clone normalization")

    if ssp.issparse(X_clone):
        X_clone = ssp.csr_matrix(X_clone)

    cell_N, clone_N = X_clone.shape
    N1, N2 = transition_map.shape
    new_coupling_matrix = ssp.lil_matrix((N1, N2))

    # cell id order in the similarity matrix is obtained by concatenating the cell id
    # list in MultiTime_cell_id_array_t1. So, we need to offset the id if we move to the next list
    offset_N1 = 0
    offset_N2 = 0
    for j in range(len(MultiTime_cell_id_array_t1)):

        logg.hint("Relative time point pair index:", j)
        cell_id_array_t1 = MultiTime_cell_id_array_t1[j]
        cell_id_array_t2 = MultiTime_cell_id_array_t2[j]

        for clone_id in range(clone_N):
            # pdb.set_trace()

            if clone_id % 1000 == 0:
                logg.hint("Clone id:", clone_id)
            idx1 = X_clone[cell_id_array_t1, clone_id].A.flatten()
            idx2 = X_clone[cell_id_array_t2, clone_id].A.flatten()
            if idx1.sum() > 0 and idx2.sum() > 0:
                ## update the new_coupling matrix
                id_1 = offset_N1 + np.nonzero(idx1)[0]
                id_2 = offset_N2 + np.nonzero(idx2)[0]
                prob = transition_map[id_1][:, id_2]

                ## try row normalization
                if normalization_mode == 0:
                    prob = hf.sparse_rowwise_multiply(
                        prob, 1 / (resol + np.sum(prob, 1))
                    )  # cell-level normalization
                else:
                    prob = prob / (
                        resol + np.sum(prob)
                    )  # clone level normalization, account for proliferation

                weight_factor = np.sqrt(
                    np.mean(idx1[idx1 > 0]) * np.mean(idx2[idx2 > 0])
                )  # the contribution of a particular clone can be tuned by its average entries
                if weight_factor > 1:
                    logg.hint(
                        "X_clone has entries not 0 or 1. Using weight modulation",
                        weight_factor,
                    )

                # Use the add mode, add up contributions from each clone
                new_coupling_matrix[id_1[:, np.newaxis], id_2] = (
                    new_coupling_matrix[id_1[:, np.newaxis], id_2]
                    + proportion[j] * prob * weight_factor
                )

        ## update offset
        offset_N1 = offset_N1 + len(cell_id_array_t1)
        offset_N2 = offset_N2 + len(cell_id_array_t2)

    ## rescale
    new_coupling_matrix = new_coupling_matrix / (new_coupling_matrix.A.max())

    ## convert to sparse matrix form
    new_coupling_matrix = new_coupling_matrix.tocsr()

    logg.hint("Start to smooth the refined clonal map")
    t = time.time()
    temp = new_coupling_matrix * final_similarity

    logg.hint("Phase I: time elapsed -- ", time.time() - t)
    smoothed_new_transition_map = initial_similarity.dot(temp)

    logg.hint("Phase II: time elapsed -- ", time.time() - t)

    # both return are numpy array
    un_SM_transition_map = new_coupling_matrix.A
    return smoothed_new_transition_map, un_SM_transition_map


def refine_Tmap_through_cospar_noSmooth(
    MultiTime_cell_id_array_t1,
    MultiTime_cell_id_array_t2,
    proportion,
    transition_map,
    X_clone,
    sparsity_threshold=0.1,
    normalization_mode=1,
):
    """
    This performs one iteration of coherent sparsity optimization

    This is the same as 'refine_Tmap_through_cospar', except that
    there is no smoothing in the end.

    Parameters
    ----------
    MultiTime_cell_id_array_t1: `np.array`
        an array of cell id sub_array, where each sub_array consists of
        clonally-related cell id's at different time points
    MultiTime_cell_id_array_t2: `np.array`
        an corresponding array of sub_array, where each sub_array are id's of
        cells that are clonally related to the corresponding sub_array at
        MultiTime_cell_id_array_t1.
    proportion: `list`
        A weight factor for each time point.
    transition_map: `np.array` or `sp.spmatrix`
        initialized transition map, or map from a previous iteration.
    X_clone: `sp.spmatrix`
        clonal matrix
    initial_similarity: `np.array`
        similarity matrix for all cells belonging
        to MultiTime_cell_id_array_t1
    final_similarity: `np.array`
        similarity matrix for all cells belonging
        to MultiTime_cell_id_array_t2
    sparsity_threshold: `float`, optional (default: 0.1)
        noise threshold to remove noises in the updated transition map,
        in the range [0,1]
    normalization_mode: `int`, optional (default: 1)
        Normalization method. Choice: [0,1].
        0, single-cell normalization; 1, Clone normalization. The clonal
        normalization suppresses the contribution of large
        clones, and is much more robust.

    Returns
    -------
    un_SM_transition_map: `np.array`
    """

    if not isinstance(X_clone[0, 0], bool):
        X_clone = X_clone.astype(bool)

    resol = 10 ** (-10)

    if normalization_mode == 0:
        logg.hint("Single-cell normalization")
    if normalization_mode == 1:
        logg.hint("Clone normalization")

    transition_map = hf.matrix_row_or_column_thresholding(
        transition_map, sparsity_threshold, row_threshold=True
    )

    if not ssp.issparse(transition_map):
        transition_map = ssp.csr_matrix(transition_map)
    if not ssp.issparse(X_clone):
        X_clone = ssp.csr_matrix(X_clone)

    cell_N, clone_N = X_clone.shape
    N1, N2 = transition_map.shape
    new_coupling_matrix = ssp.lil_matrix((N1, N2))

    offset_N1 = 0
    offset_N2 = 0
    for j in range(len(MultiTime_cell_id_array_t1)):

        logg.hint("Relative time point pair index:", j)
        cell_id_array_t1 = MultiTime_cell_id_array_t1[j]
        cell_id_array_t2 = MultiTime_cell_id_array_t2[j]

        for clone_id in range(clone_N):

            if clone_id % 1000 == 0:
                logg.hint("Clone id:", clone_id)
            idx1 = X_clone[cell_id_array_t1, clone_id].A.flatten()
            idx2 = X_clone[cell_id_array_t2, clone_id].A.flatten()
            if idx1.sum() > 0 and idx2.sum() > 0:
                ## update the new_coupling matrix
                id_1 = offset_N1 + np.nonzero(idx1)[0]
                id_2 = offset_N2 + np.nonzero(idx2)[0]
                prob = transition_map[id_1][:, id_2].A

                ## try row normalization
                if normalization_mode == 0:
                    prob = hf.sparse_rowwise_multiply(
                        prob, 1 / (resol + np.sum(prob, 1))
                    )  # cell-level normalization
                else:
                    prob = prob / (
                        resol + np.sum(prob)
                    )  # clone level normalization, account for proliferation

                weight_factor = np.sqrt(
                    np.mean(idx1[idx1 > 0]) * np.mean(idx2[idx2 > 0])
                )  # the contribution of a particular clone can be tuned by its average entries
                if weight_factor > 1:
                    logg.hint("marker gene weight", weight_factor)

                # Use the add mode, add up contributions from each clone
                new_coupling_matrix[id_1[:, np.newaxis], id_2] = (
                    new_coupling_matrix[id_1[:, np.newaxis], id_2]
                    + proportion[j] * prob * weight_factor
                )

        ## update offset
        offset_N1 = offset_N1 + len(cell_id_array_t1)
        offset_N2 = offset_N2 + len(cell_id_array_t2)

    ## convert to sparse matrix form
    new_coupling_matrix = new_coupling_matrix.tocsr()
    #
    un_SM_transition_map = new_coupling_matrix
    return un_SM_transition_map


def infer_Tmap_from_multitime_clones_private(
    adata,
    smooth_array=[15, 10, 5],
    neighbor_N=20,
    sparsity_threshold=0.1,
    intraclone_threshold=0.05,
    normalization_mode=1,
    save_subset=True,
    use_full_Smatrix=True,
    trunca_threshold=[0.001, 0.01],
    compute_new_Smatrix=False,
    max_iter_N=5,
    epsilon_converge=0.05,
):
    """
    Internal function for Tmap inference from multi-time clonal data.

    Same as :func:`.infer_Tmap_from_multitime_clones` except that it
    assumes that the adata object has been prepared for targeted
    time points. It generates the similarity matrix
    via :func:`.tmap_util.generate_similarity_matrix`, and iteratively calls
    the core function :func:`.refine_Tmap_through_cospar` to update
    the transition map.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Should be prepared by :func:`cospar.tmap._utils.select_time_points`
    smooth_array: `list`, optional (default: [15,10,5])
        List of smooth rounds at initial runs of iteration.
        Suppose that it has a length N. For iteration n<N, the n-th entry of
        smooth_array determines the kernel exponent to build the S matrix at the n-th
        iteration. When n>N, we use the last entry of smooth_array to compute
        the S matrix. We recommend starting with more smoothing depth and gradually
        reduce the depth, as inspired by simulated annealing. Data with higher
        clonal dispersion should start with higher smoothing depth. The final depth should
        depend on the manifold itself. For fewer cells, it results in a small KNN graph,
        and a small final depth should be used. We recommend to use a number at
        the multiple of 5 for computational efficiency i.e.,
        smooth_array=[20, 15, 10, 5], or [20,15,10]
    max_iter_N: `int`, optional (default: 5)
        The maximum iterations used to compute the transition map, regardless of epsilon_converge.
    epsilon_converge: `float`, optional (default: 0.05)
        The convergence threshold for the change of map correlations between consecutive iterations.
        This convergence test is activated only when CoSpar has iterated for 3 times.
    neighbor_N: `int`, optional (default: 20)
        The number of neighbors for KNN graph used for computing the similarity matrix.
    trunca_threshold: `list`, optional (default: [0.001,0.01])
        Threshold to reset entries of a matrix to zero. The first entry is for
        Similarity matrix; the second entry is for the Tmap.
        This is only for computational and storage efficiency.
    sparsity_threshold: `float`, optional (default: 0.1)
        The relative threshold to remove noises in the updated transition map,
        in the range [0,1].
    intraclone_threshold: `float`, optional (default: 0.05)
        The threshold to remove noises in the demultiplexed (un-smoothed) map,
        in the range [0,1]
    normalization_mode: `int`, optional (default: 1)
        Normalization method. Choice: [0,1].
        0, single-cell normalization; 1, Clone normalization. The clonal
        normalization suppresses the contribution of large
        clones, and is much more robust.
    save_subset: `bool`, optional (default: True)
        If true, save only Smatrix at smooth round [5,10,15,...];
        Otherwise, save Smatrix at each round.
    use_full_Smatrix: `bool`, optional (default: True)
        If true, extract the relevant Smatrix from the full Smatrix defined by all cells.
        This tends to be more accurate. The package is optimized around this choice.
    compute_new_Smatrix: `bool`, optional (default: False)
        If True, compute Smatrix from scratch, whether it was
        computed and saved before or not. This is activated only when
        `use_full_Smatrix=False`.

    Returns
    -------
    None. Inferred transition map updated at adata.uns['transition_map']
    and adata.uns['intraclone_transition_map']
    """

    ########## extract data
    clone_annot = adata.obsm["X_clone"]
    clonal_cell_id_t1 = adata.uns["clonal_cell_id_t1"]
    clonal_cell_id_t2 = adata.uns["clonal_cell_id_t2"]
    Tmap_cell_id_t1 = adata.uns["Tmap_cell_id_t1"]
    Tmap_cell_id_t2 = adata.uns["Tmap_cell_id_t2"]
    sp_idx = adata.uns["sp_idx"]
    data_des = adata.uns["data_des"][0]  # original label
    data_des_1 = adata.uns["data_des"][
        -1
    ]  # current label, sensitive to selected time points
    multiTime_cell_id_t1 = adata.uns["multiTime_cell_id_t1"]
    multiTime_cell_id_t2 = adata.uns["multiTime_cell_id_t2"]
    proportion = adata.uns["proportion"]
    data_path = settings.data_path

    ######### check whether we need to extend the map space
    ratio_t1 = np.sum(np.in1d(Tmap_cell_id_t1, clonal_cell_id_t1)) / len(
        Tmap_cell_id_t1
    )
    ratio_t2 = np.sum(np.in1d(Tmap_cell_id_t2, clonal_cell_id_t2)) / len(
        Tmap_cell_id_t2
    )
    if (ratio_t1 == 1) and (ratio_t2 == 1):
        extend_Tmap_space = False  # no need to extend the map space
    else:
        extend_Tmap_space = True

    ########################### Compute the transition map
    temp_str = "0" + str(trunca_threshold[0])[2:]

    if use_full_Smatrix:
        similarity_file_name = os.path.join(
            data_path,
            f"{data_des}_Similarity_matrix_with_all_cell_states_kNN{neighbor_N}_Truncate{temp_str}",
        )
        for round_of_smooth in smooth_array:
            if not os.path.exists(similarity_file_name + f"_SM{round_of_smooth}.npz"):
                raise ValueError(
                    f"Similarity matrix at given parameters have not been computed before! File name: {similarity_file_name}\n"
                    "Please re-run the function with: compute_new=True. If you want to use smooth round not the multiples of 5, set save_subset=False"
                )

    else:
        similarity_file_name = os.path.join(
            data_path,
            f"{data_des_1}_Similarity_matrix_with_selected_states_kNN{neighbor_N}_Truncate{temp_str}",
        )

    initial_similarity_array = []
    final_similarity_array = []
    initial_similarity_array_ext = []
    final_similarity_array_ext = []

    logg.info("Load pre-computed similarity matrix")

    if "Smatrix" not in adata.uns.keys():
        logg.hint("Load from hard disk--------")
        for round_of_smooth in smooth_array:
            # we cannot force it to compute new at this time. Otherwise, if we use_full_Smatrix, the resulting similarity is actually from adata, thus not full similarity.

            re_compute = (not use_full_Smatrix) and (
                compute_new_Smatrix
            )  # re-compute only when not using full similarity
            similarity_matrix_full = tmap_util.generate_similarity_matrix(
                adata,
                similarity_file_name,
                round_of_smooth=round_of_smooth,
                neighbor_N=neighbor_N,
                truncation_threshold=trunca_threshold[0],
                save_subset=save_subset,
                compute_new_Smatrix=re_compute,
            )

            ## add dimensionality check
            if (similarity_matrix_full.shape[0] != len(sp_idx)) or (
                similarity_matrix_full.shape[1] != len(sp_idx)
            ):
                raise ValueError(
                    "The pre-computed similarity matrix does not have the right dimension.\n"
                    "Possible reason: this is computed from a different dataset, but with the same data label in adata.uns['data_des']\n"
                    "You can fix this issue by running the Tmap inference with compute_new=True, which compute everything from scratch.\n"
                    "To avoid this from happening again, please use a different data_des for each new data, which can be set with adata.uns['data_des']=['Your_data_des']."
                )

            if use_full_Smatrix:
                # pdb.set_trace()
                similarity_matrix_full_sp = similarity_matrix_full[sp_idx][:, sp_idx]

                ### minimum similarity matrix that only involves the multi-time clones
                initial_similarity = tmap_util.generate_initial_similarity(
                    similarity_matrix_full_sp, clonal_cell_id_t1, clonal_cell_id_t1
                )
                final_similarity = tmap_util.generate_final_similarity(
                    similarity_matrix_full_sp, clonal_cell_id_t2, clonal_cell_id_t2
                )

                if extend_Tmap_space:
                    initial_similarity_ext = tmap_util.generate_initial_similarity(
                        similarity_matrix_full_sp, Tmap_cell_id_t1, clonal_cell_id_t1
                    )
                    final_similarity_ext = tmap_util.generate_final_similarity(
                        similarity_matrix_full_sp, clonal_cell_id_t2, Tmap_cell_id_t2
                    )

            else:
                initial_similarity = tmap_util.generate_initial_similarity(
                    similarity_matrix_full, clonal_cell_id_t1, clonal_cell_id_t1
                )
                final_similarity = tmap_util.generate_final_similarity(
                    similarity_matrix_full, clonal_cell_id_t2, clonal_cell_id_t2
                )

                if extend_Tmap_space:
                    initial_similarity_ext = tmap_util.generate_initial_similarity(
                        similarity_matrix_full, Tmap_cell_id_t1, clonal_cell_id_t1
                    )
                    final_similarity_ext = tmap_util.generate_final_similarity(
                        similarity_matrix_full, clonal_cell_id_t2, Tmap_cell_id_t2
                    )

            initial_similarity_array.append(initial_similarity)
            final_similarity_array.append(final_similarity)
            if extend_Tmap_space:
                initial_similarity_array_ext.append(initial_similarity_ext)
                final_similarity_array_ext.append(final_similarity_ext)

        # loading the map is too costly. We attach it to adata, and remove that after Tmap inference
        # This is useful only for the joint optimization.
        adata.uns["Smatrix"] = {}
        adata.uns["Smatrix"]["initial_similarity_array"] = initial_similarity_array
        adata.uns["Smatrix"]["final_similarity_array"] = final_similarity_array
        adata.uns["Smatrix"][
            "initial_similarity_array_ext"
        ] = initial_similarity_array_ext
        adata.uns["Smatrix"]["final_similarity_array_ext"] = final_similarity_array_ext

    else:
        logg.hint("Copy from adata (pre-loaded)--------")
        initial_similarity_array = adata.uns["Smatrix"]["initial_similarity_array"]
        final_similarity_array = adata.uns["Smatrix"]["final_similarity_array"]
        initial_similarity_array_ext = adata.uns["Smatrix"][
            "initial_similarity_array_ext"
        ]
        final_similarity_array_ext = adata.uns["Smatrix"]["final_similarity_array_ext"]

    #### Compute the core of the transition map that involve multi-time clones, then extend to other cell states
    transition_map = np.ones((len(clonal_cell_id_t1), len(clonal_cell_id_t2)))
    # transition_map_array=[transition_map_v1]

    X_clone = clone_annot.copy()
    if not ssp.issparse(X_clone):
        X_clone = ssp.csr_matrix(X_clone)

    # smooth_iter_N=len(smooth_array)
    for j in range(max_iter_N):

        # transition_map=Tmap_temp
        if j < len(smooth_array):

            logg.info(f"Iteration {j+1}, Use smooth_round={smooth_array[j]}")
            used_initial_similarity = initial_similarity_array[j]
            used_final_similarity = final_similarity_array[j]
        else:

            logg.info(f"Iteration {j+1}, Use smooth_round={smooth_array[-1]}")
            used_initial_similarity = initial_similarity_array[-1]
            used_final_similarity = final_similarity_array[-1]

        transition_map_new, unSM_sc_coupling = refine_Tmap_through_cospar(
            multiTime_cell_id_t1,
            multiTime_cell_id_t2,
            proportion,
            transition_map,
            X_clone,
            used_initial_similarity,
            used_final_similarity,
            sparsity_threshold=sparsity_threshold,
            normalization_mode=normalization_mode,
        )

        ########################### Convergency test
        # sample cell states to convergence test
        sample_N_x = 50
        sample_N_y = 100
        t0 = time.time()
        cell_N_tot_x = transition_map.shape[0]
        if cell_N_tot_x < sample_N_x:
            sample_id_temp_x = np.arange(cell_N_tot_x)
        else:
            xx = np.arange(cell_N_tot_x)
            yy = (
                list(np.nonzero(xx % 3 == 0)[0])
                + list(np.nonzero(xx % 3 == 1)[0])
                + list(np.nonzero(xx % 3 == 2)[0])
            )
            sample_id_temp_x = yy[:sample_N_x]

        cell_N_tot_y = transition_map.shape[1]
        if cell_N_tot_y < sample_N_y:
            sample_id_temp_y = np.arange(cell_N_tot_y)
        else:
            xx = np.arange(cell_N_tot_y)
            yy = (
                list(np.nonzero(xx % 3 == 0)[0])
                + list(np.nonzero(xx % 3 == 1)[0])
                + list(np.nonzero(xx % 3 == 2)[0])
            )
            sample_id_temp_y = yy[:sample_N_y]

        # transition_map is changed bytmap_core.refine_Tmap_through_cospar (thresholding). So, we only use transition_map_new to update
        if j == 0:
            X_map_0 = transition_map[sample_id_temp_x, :][:, sample_id_temp_y]
        else:
            X_map_0 = X_map_1.copy()

        X_map_1 = transition_map_new[sample_id_temp_x, :][:, sample_id_temp_y].copy()
        transition_map = transition_map_new

        if (j >= 2) and (
            j + 1 >= len(smooth_array)
        ):  # only perform convergency test after at least 3 iterations
            verbose = logg._settings_verbosity_greater_or_equal_than(3)
            corr_X = np.diag(hf.corr2_coeff(X_map_0, X_map_1)).mean()
            if verbose:
                from matplotlib import pyplot as plt

                fig = plt.figure()
                ax = plt.subplot(1, 1, 1)
                ax.plot(X_map_0.flatten(), X_map_1.flatten(), ".r")
                ax.set_xlabel("$T_{ij}$: previous iteration")
                ax.set_ylabel("$T_{ij}$: current iteration")
                ax.set_title(f"CoSpar, iter_N={j+1}, R={int(100*corr_X)/100}")
                plt.show()
            else:
                logg.info(
                    f"Convergence (CoSpar, iter_N={j+1}): corr(previous_T, current_T)={int(1000*corr_X)/1000}"
                )
                # logg.info(f"Convergence (CoSpar, iter_N={j+1}): corr(previous_T, current_T)={corr_X}; cost time={time.time()-t0}")

            if (1 - corr_X) < epsilon_converge:
                break
        #############################

    ### expand the map to other cell states
    if not extend_Tmap_space:

        logg.hint(
            "No need for Final Smooth (i.e., clonally-labeled states are the final state space for Tmap)"
        )

        transition_map = hf.matrix_row_or_column_thresholding(
            transition_map, threshold=trunca_threshold[1], row_threshold=True
        )
        adata.uns["transition_map"] = ssp.csr_matrix(transition_map)
    else:

        logg.hint(
            "Final round of Smooth (to expand the state space of Tmap to include non-clonal states)"
        )

        if j < len(smooth_array):
            used_initial_similarity_ext = initial_similarity_array_ext[j]
            used_final_similarity_ext = final_similarity_array_ext[j]
        else:
            used_initial_similarity_ext = initial_similarity_array_ext[-1]
            used_final_similarity_ext = final_similarity_array_ext[-1]

        unSM_sc_coupling = ssp.csr_matrix(unSM_sc_coupling)
        t = time.time()
        temp = unSM_sc_coupling * used_final_similarity_ext

        logg.hint("Phase I: time elapsed -- ", time.time() - t)
        transition_map_1 = used_initial_similarity_ext.dot(temp)

        logg.hint("Phase II: time elapsed -- ", time.time() - t)

        transition_map_1 = hf.matrix_row_or_column_thresholding(
            transition_map_1, threshold=trunca_threshold[1], row_threshold=True
        )
        adata.uns["transition_map"] = ssp.csr_matrix(transition_map_1)

    logg.hint("----Intraclone transition map----")

    demultiplexed_map_0 = refine_Tmap_through_cospar_noSmooth(
        multiTime_cell_id_t1,
        multiTime_cell_id_t2,
        proportion,
        transition_map,
        X_clone,
        sparsity_threshold=intraclone_threshold,
        normalization_mode=normalization_mode,
    )

    idx_t1 = hf.converting_id_from_fullSpace_to_subSpace(
        clonal_cell_id_t1, Tmap_cell_id_t1
    )[0]
    idx_t2 = hf.converting_id_from_fullSpace_to_subSpace(
        clonal_cell_id_t2, Tmap_cell_id_t2
    )[0]
    demultiplexed_map = np.zeros((len(Tmap_cell_id_t1), len(Tmap_cell_id_t2)))
    demultiplexed_map[idx_t1[:, np.newaxis], idx_t2] = demultiplexed_map_0.A
    adata.uns["intraclone_transition_map"] = ssp.csr_matrix(demultiplexed_map)


## This is just used for testing WOT
def infer_Tmap_from_optimal_transport_v0(
    adata,
    OT_epsilon=0.02,
    OT_dis_KNN=None,
    OT_solver=None,
    OT_cost=None,
    compute_new=None,
    use_existing_KNN_graph=None,
):
    """
    Test WOT

    In order to use non-uniform cell_grwoth_rate,
    we implicitly assume that you should pass a variable
    to adata at .obs['cell_growth_rate']

    Also note that only the variable `adata`, `OT_epsilon`, and .obs['cell_growth_rate']
    affects the result.

    Returns
    -------
    None. Results are stored at adata.uns['OT_transition_map'].
    """

    cell_id_array_t1 = adata.uns["Tmap_cell_id_t1"]
    cell_id_array_t2 = adata.uns["Tmap_cell_id_t2"]
    data_des = adata.uns["data_des"][0]
    data_path = settings.data_path

    logg.warn("-------------Using WOT----------------")
    logg.warn(f"epsilon={OT_epsilon}")
    import wot

    time_info = np.zeros(adata.shape[0])
    time_info[cell_id_array_t1] = 1
    time_info[cell_id_array_t2] = 2
    adata.obs["day"] = time_info

    if "cell_growth_rate" not in adata.obs.keys():
        print("Use uniform growth rate")
        adata.obs["cell_growth_rate"] = np.ones(len(time_info))
    else:
        x0 = np.mean(adata.obs["cell_growth_rate"])
        y0 = np.std(adata.obs["cell_growth_rate"])
        print(f"Use pre-supplied cell_grwoth_rate (mean: {x0:.2f}; std {y0:.2f})")

    ot_model = wot.ot.OTModel(adata, epsilon=OT_epsilon, lambda1=1, lambda2=50)
    OT_transition_map = ot_model.compute_transport_map(1, 2).X

    adata.uns["OT_transition_map"] = ssp.csr_matrix(OT_transition_map)


########### v1, with convergence test, 20210326
# We tested that, for clones of all different sizes, where np.argsort gives unique results,
# this method reproduces the v01, v1 results, when use_fixed_clonesize_t1=True, and when change
# sort_clone=0,1,-1.
def refine_Tmap_through_joint_optimization(
    adata,
    initialized_map,
    smooth_array=[15, 10, 5],
    max_iter_N=[1, 5],
    epsilon_converge=[0.05, 0.05],
    CoSpar_KNN=20,
    normalization_mode=1,
    sparsity_threshold=0.2,
    use_full_Smatrix=True,
    trunca_threshold=[0.001, 0.01],
    compute_new=True,
    use_fixed_clonesize_t1=False,
    sort_clone=1,
    save_subset=True,
):
    """
    Infer Tmap from clones with a single time point

    Starting from an initialized transitin map from state information,
    we jointly infer the initial clonal states and the transition map.

    This method has been optimized to be very fast. Besides, it is
    deterministic.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Should have only two time points.
    initialized_map: `sp.spmatrix`
        Initialized transition map based on state information alone.
    normalization_mode: `int`, optional (default: 1)
        Normalization method. Choice: [0,1].
        0, single-cell normalization; 1, Clone normalization. The clonal
        normalization suppresses the contribution of large
        clones, and is much more robust.
    smooth_array: `list`, optional (default: [15,10,5])
        List of smooth rounds at initial runs of iteration.
        Suppose that it has a length N. For iteration n<N, the n-th entry of
        smooth_array determines the kernel exponent to build the S matrix at the n-th
        iteration. When n>N, we use the last entry of smooth_array to compute
        the S matrix. We recommend starting with more smoothing depth and gradually
        reduce the depth, as inspired by simulated annealing. Data with higher
        clonal dispersion should start with higher smoothing depth. The final depth should
        depend on the manifold itself. For fewer cells, it results in a small KNN graph,
        and a small final depth should be used. We recommend to use a number at
        the multiple of 5 for computational efficiency i.e.,
        smooth_array=[20, 15, 10, 5], or [20,15,10]
    max_iter_N: `list`, optional (default: [1,5])
        A list for maximum iterations for the Joint optimization and CoSpar core function, respectively.
    epsilon_converge: `list`, optional (default: [0.05,0.05])
        A list of convergence threshold for the Joint optimization and CoSpar core function, respectively.
        The convergence threshold is for the change of map correlations between consecutive iterations.
        For CoSpar core function, this convergence test is activated only when CoSpar has iterated for 3 times.
    CoSpar_KNN: `int`, optional (default: 20)
        The number of neighbors for KNN graph used for computing the similarity matrix.
    trunca_threshold: `list`, optional (default: [0.001,0.01])
        Threshold to reset entries of a matrix to zero. The first entry is for
        Similarity matrix; the second entry is for the Tmap.
        This is only for computational and storage efficiency.
    sparsity_threshold: `float`, optional (default: 0.1)
        The relative threshold to remove noises in the updated transition map,
        in the range [0,1].
    save_subset: `bool`, optional (default: True)
        If true, save only Smatrix at smooth round [5,10,15,...];
        Otherwise, save Smatrix at each round.
    use_full_Smatrix: `bool`, optional (default: True)
        If true, extract the relevant Smatrix from the full Smatrix defined by all cells.
        This tends to be more accurate. The package is optimized around this choice.
    use_fixed_clonesize_t1: `bool`, optional (default: False)
        If true, fix the number of initial states as the same for all clones
    sort_clone: `int`, optional (default: 1)
        The order to infer initial states for each clone: {1,-1,others}.
        1, sort clones by size from small to large;
        -1, sort clones by size from large to small;
        others, do not sort.
    compute_new: `bool`, optional (default: False)
        If True, compute everything (ShortestPathDis, OT_map, etc.) from scratch,
        whether it was computed and saved before or not.
    save_subset: `bool`, optional (default: True)
        If true, save only Smatrix at smooth round [5,10,15,...];
        Otherwise, save Smatrix at each round.

    Returns
    ------
    None. Update adata.obsm['X_clone'] and adata.uns['transition_map'],
    as well as adata.uns['OT_transition_map'] or
    adata.uns['HighVar_transition_map'], depending on the initialization.
    """

    # I found the error: 1) we should use clonally related cell number at t2 as a factor to determine the clonally cell number at t1
    #                    2) update the whole t2 clonal info at once

    # logg.info("Joint optimization that consider possibility of clonal overlap")

    cell_id_array_t1 = adata.uns["Tmap_cell_id_t1"]
    cell_id_array_t2 = adata.uns["Tmap_cell_id_t2"]
    data_des = adata.uns["data_des"][-1]
    data_path = settings.data_path
    X_clone = adata.obsm["X_clone"]
    if not ssp.issparse(X_clone):
        X_clone = ssp.csr_matrix(X_clone)

    time_info = np.array(adata.obs["time_info"])
    time_index_t1 = time_info == (time_info[cell_id_array_t1[0]])
    time_index_t2 = time_info == (time_info[cell_id_array_t2[0]])

    if not ssp.issparse(initialized_map):
        map_temp = ssp.csr_matrix(initialized_map)
    else:
        map_temp = initialized_map

    # a clone must has at least 2 cells, to be updated later.
    valid_clone_id = np.nonzero(X_clone[cell_id_array_t2].sum(0).A.flatten() > 0)[0]
    X_clone_temp = X_clone[:, valid_clone_id]
    clonal_cells_t2 = np.sum(X_clone_temp[cell_id_array_t2].sum(1).flatten())

    logg.hint(f"original clone shape: {X_clone.shape}")
    logg.hint(f"After excluding zero-sized clones at t2: {X_clone_temp.shape}")

    flag = True  # to check whether overlapping clones are found or not
    if use_fixed_clonesize_t1:
        logg.hint("Use fixed clone size at t1")

    ##### Partition cells into non-overlapping, combinatorial BC_id.
    # ---------------------------------
    # find the combinatorial barcodes
    clone_idx = np.nonzero(X_clone_temp.A)
    dic = [[] for j in range(X_clone_temp.shape[0])]  # a list of list
    for j in range(clone_idx[0].shape[0]):
        dic[clone_idx[0][j]].append(clone_idx[1][j])

    BC_id = [
        tuple(x) for x in dic
    ]  # a BC_id is a unique barcode combination, does not change the ordering of cells

    # --------------------
    # construct the new X_clone_temp matrix, and the clone_mapping
    unique_BC_id = list(set(BC_id))
    if () in unique_BC_id:  # () is resulted from cells without any barcodes
        unique_BC_id.remove(())

    # construct a X_clone_newBC for the new BC_id
    # also record how the new BC_id is related to the old barcode

    X_clone_newBC = np.zeros((X_clone_temp.shape[0], len(unique_BC_id)))
    for i, BC_0 in enumerate(BC_id):
        for j, BC_1 in enumerate(unique_BC_id):
            if BC_1 == BC_0:
                X_clone_newBC[i, j] = 1  # does not change the ordering of cells

    clone_mapping = np.zeros((X_clone_temp.shape[1], X_clone_newBC.shape[1]))
    for j, BC_1 in enumerate(unique_BC_id):
        for k in BC_1:
            clone_mapping[k, j] = 1

    X_clone_newBC = ssp.csr_matrix(X_clone_newBC)
    clone_mapping = ssp.csr_matrix(clone_mapping)
    # To recover the original X_clone_temp, use 'X_clone_newBC*(clone_mapping.T)'
    # howver, clone_mapping is not invertible. We cannot get from X_clone_temp to
    # X_clone_newBC using matrix multiplification.

    ### select the early states using the grouped distribution of a clone
    ### clones are not overlapping, and all early states should be attached to clones at the end

    # we sort clones according to their sizes. The order of cells are not affected. So, it should not affect downstream analysis
    # small clones tend to be the ones that are barcoded/mutated later, while large clones tend to be early mutations...
    clone_size_t2_temp = X_clone_newBC[cell_id_array_t2].sum(0).A.flatten()

    if sort_clone == 1:
        logg.hint("Sort clones by size (small to large)")

        sort_clone_id = np.argsort(clone_size_t2_temp, kind="stable")
        clone_size_t2 = clone_size_t2_temp[sort_clone_id]
        X_clone_sort = X_clone_newBC[:, sort_clone_id]
        clone_mapping_sort = clone_mapping[:, sort_clone_id]

    elif sort_clone == -1:
        logg.hint("Sort clones by size (large to small)")

        sort_clone_id = np.argsort(clone_size_t2_temp, kind="stable")[::-1]
        clone_size_t2 = clone_size_t2_temp[sort_clone_id]
        X_clone_sort = X_clone_newBC[:, sort_clone_id]
        clone_mapping_sort = clone_mapping[:, sort_clone_id]

    else:
        logg.hint("Do not order clones by size ")
        clone_size_t2 = clone_size_t2_temp
        X_clone_sort = X_clone_newBC
        clone_mapping_sort = clone_mapping

    logg.hint("Infer the number of initial cells to extract for each clone in advance")
    clone_N1 = X_clone_sort.shape[1]
    ave_clone_size_t1 = int(np.ceil(len(cell_id_array_t1) / clone_N1))
    cum_cell_N = np.ceil(
        np.cumsum(clone_size_t2) * len(cell_id_array_t1) / clonal_cells_t2
    )
    cell_N_to_extract = np.zeros(len(cum_cell_N), dtype=int)
    if use_fixed_clonesize_t1:
        cell_N_to_extract += ave_clone_size_t1
    else:
        cell_N_to_extract[0] = cum_cell_N[0]
        cell_N_to_extract[1:] = np.diff(cum_cell_N)

    for x0 in range(max_iter_N[0]):
        logg.info(f"-----JointOpt Iteration {x0+1}: Infer initial clonal structure")

        # update initial state probability matrix based on the current map
        initial_prob_matrix = (
            map_temp * X_clone_sort[cell_id_array_t2]
        ).A  # a initial probability matrix for t1 cells, shape (n_t1_cell,n_clone)

        ########## begin: update clones
        remaining_ids_t1 = list(np.arange(len(cell_id_array_t1), dtype=int))

        X_clone_new = np.zeros(X_clone_sort.shape, dtype=bool)
        X_clone_new[cell_id_array_t2] = X_clone_sort[cell_id_array_t2].A.astype(
            bool
        )  # update the whole t2 clones at once

        for j in range(clone_N1):
            if j % 100 == 0:
                # pdb.set_trace()
                logg.hint(f"Inferring early clonal states: current clone id {j}")

            # infer the earlier clonal states for each clone
            ### select the early states using the grouped distribution of a clone
            sorted_id_array = np.argsort(
                initial_prob_matrix[remaining_ids_t1, j], kind="stable"
            )[::-1]

            sel_id_t1 = sorted_id_array[: cell_N_to_extract[j]]
            temp_t1_idx = np.zeros(len(cell_id_array_t1), dtype=bool)
            temp_t1_idx[np.array(remaining_ids_t1)[sel_id_t1]] = True
            X_clone_new[cell_id_array_t1, j] = temp_t1_idx
            for kk in np.array(remaining_ids_t1)[sel_id_t1]:
                remaining_ids_t1.remove(kk)

            if (len(remaining_ids_t1) == 0) and ((j + 1) < clone_N1):
                logg.hint(f"Early break; current clone id: {j+1}")
                break

        ########### end: update clones
        cell_id_array_t1_new = np.nonzero((X_clone_new.sum(1) > 0) & (time_index_t1))[0]
        cell_id_array_t2_new = np.nonzero((X_clone_new.sum(1) > 0) & (time_index_t2))[0]

        adata.obsm["X_clone"] = ssp.csr_matrix(X_clone_new) * (
            clone_mapping_sort.T
        )  # convert back to the original clone structure
        adata.uns["multiTime_cell_id_t1"] = [
            cell_id_array_t1_new
        ]  # For CoSpar, clonally-related states
        adata.uns["multiTime_cell_id_t2"] = [cell_id_array_t2_new]
        adata.uns[
            "clonal_cell_id_t1"
        ] = cell_id_array_t1_new  # for prepare the similarity matrix with same cell states
        adata.uns["clonal_cell_id_t2"] = cell_id_array_t2_new
        adata.uns["proportion"] = [1]

        logg.info(
            f"-----JointOpt Iteration {x0+1}: Update the transition map by CoSpar"
        )
        infer_Tmap_from_multitime_clones_private(
            adata,
            smooth_array=smooth_array,
            neighbor_N=CoSpar_KNN,
            sparsity_threshold=sparsity_threshold,
            normalization_mode=normalization_mode,
            save_subset=save_subset,
            use_full_Smatrix=use_full_Smatrix,
            trunca_threshold=trunca_threshold,
            compute_new_Smatrix=compute_new,
            max_iter_N=max_iter_N[1],
            epsilon_converge=epsilon_converge[1],
        )

        # update, for the next iteration
        if "transition_map" in adata.uns.keys():

            # sample cell states to perform the accuracy test
            sample_N_x = 50
            sample_N_y = 100
            t0 = time.time()
            cell_N_tot_x = map_temp.shape[0]
            if cell_N_tot_x < sample_N_x:
                sample_id_temp_x = np.arange(cell_N_tot_x)
            else:
                xx = np.arange(cell_N_tot_x)
                yy = (
                    list(np.nonzero(xx % 3 == 0)[0])
                    + list(np.nonzero(xx % 3 == 1)[0])
                    + list(np.nonzero(xx % 3 == 2)[0])
                )
                sample_id_temp_x = yy[:sample_N_x]

            cell_N_tot_y = map_temp.shape[1]
            if cell_N_tot_y < sample_N_y:
                sample_id_temp_y = np.arange(cell_N_tot_y)
            else:
                xx = np.arange(cell_N_tot_y)
                yy = (
                    list(np.nonzero(xx % 3 == 0)[0])
                    + list(np.nonzero(xx % 3 == 1)[0])
                    + list(np.nonzero(xx % 3 == 2)[0])
                )
                sample_id_temp_y = yy[:sample_N_y]

            if x0 == 0:
                X_map_0 = map_temp[sample_id_temp_x, :][:, sample_id_temp_y].A
            else:
                X_map_0 = X_map_1.copy()

            X_map_1 = adata.uns["transition_map"][sample_id_temp_x, :][
                :, sample_id_temp_y
            ].A

            verbose = logg._settings_verbosity_greater_or_equal_than(3)
            corr_X = np.diag(hf.corr2_coeff(X_map_0, X_map_1)).mean()
            if verbose:
                from matplotlib import pyplot as plt

                fig = plt.figure()
                ax = plt.subplot(1, 1, 1)
                ax.plot(X_map_0.flatten(), X_map_1.flatten(), ".r")
                ax.set_xlabel("$T_{ij}$: previous iteration")
                ax.set_ylabel("$T_{ij}$: current iteration")
                ax.set_title(f"Joint Opt., iter_N={x0+1}, R={int(100*corr_X)/100}")
                plt.show()
            else:
                # logg.info(f"Convergence (JointOpt, iter_N={x0+1}): corr(previous_T, current_T)={int(1000*corr_X)/1000}; cost time={time.time()-t0}")
                logg.info(
                    f"Convergence (JointOpt, iter_N={x0+1}): corr(previous_T, current_T)={int(1000*corr_X)/1000}"
                )

            if abs(1 - corr_X) < epsilon_converge[0]:
                break

            map_temp = adata.uns["transition_map"]
        else:
            raise ValueError(
                "transition_map not updated in infer_Tmap_from_multitime_clones_private."
            )


def infer_Tmap_from_HighVar(
    adata,
    min_counts=3,
    min_cells=3,
    min_gene_vscore_pctl=85,
    smooth_array=[15, 10, 5],
    neighbor_N=20,
    sparsity_threshold=0.2,
    normalization_mode=1,
    use_full_Smatrix=True,
    trunca_threshold=[0.001, 0.01],
    compute_new_Smatrix=True,
    max_iter_N=5,
    epsilon_converge=0.05,
    save_subset=True,
):
    """
    Generate Tmap based on state information using HighVar.

    We convert differentially expressed genes into `pseudo-clones`,
    and run coherent sparsity optimization to infer the transition map.
    Each clone occupies a different set of cells.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assumed to be preprocessed, only has two time points.
    min_counts: int, optional (default: 3)
        Minimum number of UMIs per cell to be considered for selecting highly variable genes.
    min_cells: int, optional (default: 3)
        Minimum number of cells per gene to be considered for selecting highly variable genes.
    min_gene_vscore_pctl: int, optional (default: 85)
        Genes with a variability percentile higher than this threshold are marked as
        highly variable genes for constructing pseudo-clones. Range: [0,100].
    smooth_array: `list`, optional (default: [15,10,5])
        List of smooth rounds at initial runs of iteration.
        Suppose that it has a length N. For iteration n<N, the n-th entry of
        smooth_array determines the kernel exponent to build the S matrix at the n-th
        iteration. When n>N, we use the last entry of smooth_array to compute
        the S matrix. We recommend starting with more smoothing depth and gradually
        reduce the depth, as inspired by simulated annealing. Data with higher
        clonal dispersion should start with higher smoothing depth. The final depth should
        depend on the manifold itself. For fewer cells, it results in a small KNN graph,
        and a small final depth should be used. We recommend to use a number at
        the multiple of 5 for computational efficiency i.e.,
        smooth_array=[20, 15, 10, 5], or [20,15,10]
    max_iter_N: `int`, optional (default: 5)
        The maximum iterations used to compute the transition map,
        regardless of epsilon_converge.
    epsilon_converge: `float`, optional (default: 0.05)
        The convergence threshold for the change of map
        correlations between consecutive iterations.
        This convergence test is activated only when
        CoSpar has iterated for 3 times.
    neighbor_N: `int`, optional (default: 20)
        The number of neighbors for KNN graph used for computing the similarity matrix.
    trunca_threshold: `list`, optional (default: [0.001,0.01])
        Threshold to reset entries of a matrix to zero. The first entry is for
        Similarity matrix; the second entry is for the Tmap.
        This is only for computational and storage efficiency.
    sparsity_threshold: `float`, optional (default: 0.1)
        The relative threshold to remove noises in the updated transition map,
        in the range [0,1].
    normalization_mode: `int`, optional (default: 1)
        Normalization method. Choice: [0,1].
        0, single-cell normalization; 1, Clone normalization. The clonal
        normalization suppresses the contribution of large
        clones, and is much more robust.
    save_subset: `bool`, optional (default: True)
        If true, save only Smatrix at smooth round [5,10,15,...];
        Otherwise, save Smatrix at each round.
    use_full_Smatrix: `bool`, optional (default: True)
        If true, extract the relevant Smatrix from the full Smatrix defined by all cells.
        This tends to be more accurate. The package is optimized around this choice.
    compute_new_Smatrix: `bool`, optional (default: False)
        If True, compute Smatrix from scratch, whether it was
        computed and saved before or not.

    Returns
    -------
    None. Results are stored at adata.uns['HighVar_transition_map'].
    """

    # logg.info("HighVar-v0: avoid cells that have been selected")
    weight = 1  # wehight of each gene.

    cell_id_array_t1 = adata.uns["Tmap_cell_id_t1"]
    cell_id_array_t2 = adata.uns["Tmap_cell_id_t2"]
    real_clone_annot = adata.obsm["X_clone"]

    time_info = np.array(adata.obs["time_info"])
    selected_time_points = [
        time_info[cell_id_array_t1][0],
        time_info[cell_id_array_t2][0],
    ]

    # logg.info("----------------")
    logg.info("Step a: find the commonly shared highly variable genes------")
    adata_t1 = sc.AnnData(adata.X[cell_id_array_t1])
    adata_t2 = sc.AnnData(adata.X[cell_id_array_t2])

    ## use marker genes
    gene_list = adata.var_names

    verbose = logg._settings_verbosity_greater_or_equal_than(3)

    ## the scanpy version
    # def filter_gene_custom(adata_tmp, gene_list_tmp):
    #     sc.pp.normalize_total(adata_tmp, target_sum=1e4)
    #     sc.pp.log1p(adata_tmp)
    #     sc.pp.highly_variable_genes(
    #         adata_tmp, min_mean=0.0125, max_mean=3, min_disp=0.5
    #     )
    #     return list(gene_list_tmp[adata_tmp.var["highly_variable"]])
    # highvar_genes_t1 = filter_gene_custom(adata_t1, gene_list)
    # highvar_genes_t2 = filter_gene_custom(adata_t2, gene_list)

    gene_idx_t1 = hf.filter_genes(
        adata_t1.X,
        min_counts=min_counts,
        min_cells=min_cells,
        min_vscore_pctl=min_gene_vscore_pctl,
        show_vscore_plot=verbose,
    )
    if gene_idx_t1 is not None:
        highvar_genes_t1 = gene_list[gene_idx_t1]
    else:
        raise ValueError("No variable genes selected at t1")

    gene_idx_t2 = hf.filter_genes(
        adata_t2.X,
        min_counts=min_counts,
        min_cells=min_cells,
        min_vscore_pctl=min_gene_vscore_pctl,
        show_vscore_plot=verbose,
    )
    if gene_idx_t2 is not None:
        highvar_genes_t2 = gene_list[gene_idx_t2]
    else:
        raise ValueError("No variable genes selected at t2")

    common_gene = sorted(list(set(highvar_genes_t1).intersection(highvar_genes_t2)))

    logg.info(
        f"Highly varable gene number: {len(highvar_genes_t1)} (t1); {len(highvar_genes_t2)} (t2). Common set: {len(common_gene)}"
    )

    # logg.info("----------------")
    logg.info("Step b: convert the shared highly variable genes into clonal info------")

    sel_marker_gene_list = common_gene.copy()
    clone_annot_gene = np.zeros((adata.shape[0], len(sel_marker_gene_list)))
    N_t1 = len(cell_id_array_t1)
    N_t2 = len(cell_id_array_t2)
    cumu_sel_idx_t1 = np.zeros(N_t1, dtype=bool)
    cumu_sel_idx_t2 = np.zeros(N_t2, dtype=bool)
    cell_fraction_per_gene = 1 / len(
        sel_marker_gene_list
    )  # fraction of cells as clonally related by this gene
    cutoff_t1 = int(np.ceil(len(cell_id_array_t1) * cell_fraction_per_gene))
    cutoff_t2 = int(np.ceil(len(cell_id_array_t2) * cell_fraction_per_gene))
    gene_exp_matrix = adata[:, sel_marker_gene_list].X.A
    for j in tqdm(range(gene_exp_matrix.shape[1])):
        temp_t1 = gene_exp_matrix[:, j][cell_id_array_t1]
        temp_t1[cumu_sel_idx_t1] = 0  # set selected cell id to have zero expression
        sel_id_t1 = np.argsort(temp_t1, kind="stable")[::-1][:cutoff_t1]
        clone_annot_gene[cell_id_array_t1[sel_id_t1], j] = weight
        cumu_sel_idx_t1[sel_id_t1] = True

        temp_t2 = gene_exp_matrix[:, j][cell_id_array_t2]
        temp_t2[cumu_sel_idx_t2] = 0  # set selected cell id to have zero expression
        sel_id_t2 = np.argsort(temp_t2, kind="stable")[::-1][:cutoff_t2]
        clone_annot_gene[cell_id_array_t2[sel_id_t2], j] = weight
        cumu_sel_idx_t2[sel_id_t2] = True

        if (np.sum(~cumu_sel_idx_t1) == 0) or (np.sum(~cumu_sel_idx_t2) == 0):
            logg.info(f"Total used genes={j} (no cells left)")
            break

    logg.info(
        "Step c: compute the transition map based on clonal info from highly variable genes------"
    )

    adata.obsm["X_clone"] = ssp.csr_matrix(clone_annot_gene)
    adata.uns["multiTime_cell_id_t1"] = [cell_id_array_t1]
    adata.uns["multiTime_cell_id_t2"] = [cell_id_array_t2]
    adata.uns["proportion"] = [1]

    infer_Tmap_from_multitime_clones_private(
        adata,
        smooth_array=smooth_array,
        neighbor_N=neighbor_N,
        sparsity_threshold=sparsity_threshold,
        normalization_mode=normalization_mode,
        save_subset=save_subset,
        use_full_Smatrix=use_full_Smatrix,
        trunca_threshold=trunca_threshold,
        compute_new_Smatrix=compute_new_Smatrix,
        max_iter_N=max_iter_N,
        epsilon_converge=epsilon_converge,
    )

    adata.uns["HighVar_transition_map"] = adata.uns["transition_map"]
    adata.obsm[
        "X_clone"
    ] = real_clone_annot  # This entry has been changed previously. Note correct the clonal matrix
    # data_des_1=data_des_0+'_HighVar1' # to record which initialization is used
    # adata.uns['data_des']=[data_des_orig,data_des_1]

    if "Smatrix" in adata.uns.keys():
        adata.uns.pop("Smatrix")


# this is the new version: v1, finally used
def infer_Tmap_from_optimal_transport(
    adata,
    OT_epsilon=0.02,
    OT_dis_KNN=5,
    OT_solver="duality_gap",
    OT_cost="SPD",
    compute_new=True,
    use_existing_KNN_graph=False,
):
    """
    Compute Tmap from state info using optimal transport (OT).

    We provide the options for the OT solver, and also the cost function.
    The OT solver does not seem to matter, although 'duality_gap' is faster.
    The cost function could affect the OT map results. Using shortest path
    distance ('SPD') is slower but more accurate, while using gene expression
    distance ('GED') is faster but less accurate. The performance of cospar
    is robust to the initialized map (this is especially so in terms of fate
    bias, not so much for the fate map alone)

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assumed to be preprocessed, only has two time points.
    OT_epsilon: `float`, optional (default: 0.02)
        The entropic regularization, >0. A larger value increases
        uncertainty of the transition.
    OT_dis_KNN: `int`, optional (default: 5)
        Number of nearest neighbors to construct the KNN graph for
        computing the shortest path distance.
    OT_solver: `str`, optional (default: `duality_gap`)
        The method used to compute the optimal transport map. Available choices:
        {'duality_gap','fixed_iters'}. Our test shows that they produce the same
        results, while 'duality_gap' is almost twice faster.
    OT_cost: `str`, optional (default: `SPD`), options {'GED','SPD'}
        The cost metric. We provide gene expression distance (GED), and also
        shortest path distance (SPD). GED is much faster, but SPD is more accurate.
        However, coherent sparsity optimization is robust to the initialization.
    compute_new: `bool`, optional (default: False)
        If True, compute OT_map and also the shortest path distance from scratch,
        whether it was computed and saved before or not.
    use_existing_KNN_graph: `bool`, optional (default: False)
        If true and adata.obsp['connectivities'], use the existing knn graph for constructing
        the shortest-path distance. This overrides all other parameters.

    Returns
    -------
    None. Results are stored at adata.uns['OT_transition_map'].
    """

    cell_id_array_t1 = adata.uns["Tmap_cell_id_t1"]
    cell_id_array_t2 = adata.uns["Tmap_cell_id_t2"]
    data_des = adata.uns["data_des"][-1]
    data_path = settings.data_path

    ######## apply optimal transport
    CustomOT_file_name = os.path.join(
        data_path,
        f"{data_des}_CustomOT_map_epsilon{OT_epsilon}_KNN{OT_dis_KNN}_OTcost{OT_cost}.npz",
    )
    if os.path.exists(CustomOT_file_name) and (not compute_new):

        logg.info("Load pre-computed custom OT matrix")
        OT_transition_map = ssp.load_npz(CustomOT_file_name)

        ## add dimensionality check
        if (OT_transition_map.shape[0] != len(cell_id_array_t1)) or (
            OT_transition_map.shape[1] != len(cell_id_array_t2)
        ):
            raise ValueError(
                "The pre-computed OT transition map does not have the right dimension.\n"
                "Possible reason: this is computed from a different dataset, but with the same data label in adata.uns['data_des']\n"
                "You can fix this issue by running the Tmap inference with compute_new=True, which compute everything from scratch.\n"
                "To avoid this from happening again, please use a different data_des for each new data, which can be set with adata.uns['data_des']=['Your_data_des']."
            )

    else:

        ############ Compute shorted-path distance
        # use sklearn KNN graph construction method and select the connectivity option, not related to UMAP
        # use the mode 'distance' to obtain the shortest-path *distance*, rather than 'connectivity'
        if OT_cost == "SPD":
            SPD_file_name = os.path.join(
                data_path,
                f"{data_des}_ShortestPathDistanceMatrix_t0t1_KNN{OT_dis_KNN}.npy",
            )
            if os.path.exists(SPD_file_name) and (not compute_new):

                logg.info("Load pre-computed shortest path distance matrix")
                OT_cost_matrix = np.load(SPD_file_name)

                ## add dimensionality check
                if (OT_cost_matrix.shape[0] != len(cell_id_array_t1)) or (
                    OT_cost_matrix.shape[1] != len(cell_id_array_t2)
                ):
                    raise ValueError(
                        "The pre-computed OT cost matrix does not have the right dimension.\n"
                        "Possible reason: this is computed from a different dataset, but with the same data label in adata.uns['data_des']\n"
                        "You can fix this issue by running the Tmap inference with compute_new=True, which compute everything from scratch.\n"
                        "To avoid this from happening again, please use a different data_des for each new data, which can be set with adata.uns['data_des']=['Your_data_des']."
                    )

            else:

                logg.info("Compute new shortest path distance matrix")
                t = time.time()
                # data_matrix=adata.obsm['X_pca']
                # ShortPath_dis=hf.compute_shortest_path_distance_from_raw_matrix(data_matrix,num_neighbors_target=OT_dis_KNN,mode='distance')
                ShortPath_dis = hf.compute_shortest_path_distance(
                    adata,
                    num_neighbors_target=OT_dis_KNN,
                    mode="distances",
                    method="umap",
                    use_existing_KNN_graph=use_existing_KNN_graph,
                )

                idx0 = cell_id_array_t1
                idx1 = cell_id_array_t2
                ShortPath_dis_t0t1 = ShortPath_dis[idx0[:, np.newaxis], idx1]
                OT_cost_matrix = ShortPath_dis_t0t1 / ShortPath_dis_t0t1.max()

                np.save(
                    SPD_file_name, OT_cost_matrix
                )  # This is not a sparse matrix at all.

                logg.info(
                    f"Finishing computing shortest-path distance, used time {time.time()-t}"
                )
        else:
            t = time.time()
            pc_n = adata.obsm["X_pca"].shape[1]
            OT_cost_matrix = hf.compute_gene_exp_distance(
                adata, cell_id_array_t1, cell_id_array_t2, pc_n=pc_n
            )
            logg.info(
                f"Finishing computing gene expression distance, used time {time.time()-t}"
            )

        ##################
        logg.info("Compute new custom OT matrix")
        t = time.time()

        if "cell_growth_rate" not in adata.obs.keys():
            logg.info("Use uniform growth rate")
            adata.obs["cell_growth_rate"] = np.ones(adata.shape[0])
        else:
            logg.info("Use pre-supplied cell_grwoth_rate")
            x0 = np.mean(adata.obs["cell_growth_rate"])
            y0 = np.std(adata.obs["cell_growth_rate"])
            print(f"Use pre-supplied cell_grwoth_rate (mean: {x0:.2f}; std {y0:.2f})")

        #############
        OT_solver = "duality_gap"
        logg.info(f"OT solver: {OT_solver}")
        if (
            OT_solver == "fixed_iters"
        ):  # This takes 50s for the subsampled hematopoietic data. The result is the same.
            ot_config = {
                "C": OT_cost_matrix,
                "G": np.array(adata.obs["cell_growth_rate"])[cell_id_array_t1],
                "epsilon": OT_epsilon,
                "lambda1": 1,
                "lambda2": 50,
                "epsilon0": 1,
                "scaling_iter": 3000,
                "tau": 10000,
                "inner_iter_max": 50,
                "extra_iter": 1000,
            }

            OT_transition_map = transport_stablev2(**ot_config)

        elif (
            OT_solver == "duality_gap"
        ):  # This takes 30s for the subsampled hematopoietic data. The result is the same.
            ot_config = {
                "C": OT_cost_matrix,
                "G": np.array(adata.obs["cell_growth_rate"])[cell_id_array_t1],
                "epsilon": OT_epsilon,
                "lambda1": 1,
                "lambda2": 50,
                "epsilon0": 1,
                "tau": 10000,
                "tolerance": 1e-08,
                "max_iter": 1e7,
                "batch_size": 5,
            }

            OT_transition_map = optimal_transport_duality_gap(**ot_config)

        else:
            raise ValueError("Unknown solver")

        OT_transition_map = hf.matrix_row_or_column_thresholding(
            OT_transition_map, threshold=0.01
        )
        if not ssp.issparse(OT_transition_map):
            OT_transition_map = ssp.csr_matrix(OT_transition_map)
        # ssp.save_npz(CustomOT_file_name, OT_transition_map)

        logg.info(
            f"Finishing computing optial transport map, used time {time.time()-t}"
        )

    adata.uns["OT_transition_map"] = OT_transition_map



import os
import time

import numpy as np
import pandas as pd
import scanpy as sc
import scipy.sparse as ssp

from .. import help_functions as hf
from .. import logging as logg
from .. import plotting as pl
from .. import settings
from .. import tool as tl


def generate_similarity_matrix(
    adata,
    file_name,
    round_of_smooth=10,
    neighbor_N=20,
    beta=0.1,
    truncation_threshold=0.001,
    save_subset=True,
    use_existing_KNN_graph=False,
    compute_new_Smatrix=False,
):
    """
    Generate similarity matrix (Smatrix) through graph diffusion

    It generates the similarity matrix via iterative graph diffusion.
    Similarity matrix from each round of diffusion will be saved, after truncation
    to promote sparsity and save space. If save_subset is activated, only save
    Smatrix for smooth rounds at the multiples of 5 (like 5,10,15,...). If a Smatrix is pre-computed,
    it will be loaded directly if compute_new_Smatrix=Flase.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    file_name: str
        Filename to load pre-computed similarity matrix or save the newly
        computed similarity matrix.
    round_of_smooth: `int`, optional (default: 10)
        The rounds of graph diffusion.
    neighbor_N: `int`, optional (default: 20)
        Neighber number for constructing the KNN graph, using the UMAP method.
    beta: `float`, option (default: 0.1)
        Probability to stay at the origin in a unit diffusion step, in the range [0,1]
    truncation_threshold: `float`, optional (default: 0.001)
        At each iteration, truncate the similarity matrix using
        truncation_threshold. This promotes the sparsity of the matrix,
        thus the speed of computation. We set the truncation threshold to be small,
        to guarantee accracy.
    save_subset: `bool`, optional (default: True)
        If true, save only Smatrix at smooth round [5,10,15,...];
        Otherwise, save Smatrix at each round.
    use_existing_KNN_graph: `bool`, optional (default: False)
        If true and adata.obsp['connectivities'], use the existing knn graph to build
        the similarity matrix, regardless of neighbor_N.
    compute_new_Smatrix: `bool`, optional (default: False)
        If true, compute a new Smatrix, even if there is pre-computed Smatrix with the
        same parameterization.

    Returns
    -------
        similarity_matrix: `sp.spmatrix`
    """

    if os.path.exists(file_name + f"_SM{round_of_smooth}.npz") and (
        not compute_new_Smatrix
    ):

        logg.hint("Compute similarity matrix: load existing data")
        similarity_matrix = ssp.load_npz(file_name + f"_SM{round_of_smooth}.npz")
    else:  # compute now

        logg.hint(f"Compute similarity matrix: computing new; beta={beta}")

        # add a step to compute PCA in case this is not computed

        if (not use_existing_KNN_graph) or ("connectivities" not in adata.obsp.keys()):
            # here, we assume that adata already has pre-computed PCA
            sc.pp.neighbors(adata, n_neighbors=neighbor_N)
        else:
            logg.hint(
                "Use existing KNN graph at adata.obsp['connectivities'] for generating the smooth matrix"
            )
        adjacency_matrix = adata.obsp["connectivities"]

        ############## The new method
        adjacency_matrix = (adjacency_matrix + adjacency_matrix.T) / 2
        ##############

        adjacency_matrix = hf.sparse_rowwise_multiply(
            adjacency_matrix, 1 / adjacency_matrix.sum(1).A.squeeze()
        )
        nrow = adata.shape[0]
        similarity_matrix = ssp.lil_matrix((nrow, nrow))
        similarity_matrix.setdiag(np.ones(nrow))
        transpose_A = adjacency_matrix.T

        if round_of_smooth == 0:
            SM = 0
            similarity_matrix = ssp.csr_matrix(similarity_matrix)
            ssp.save_npz(file_name + f"_SM{SM}.npz", similarity_matrix)

        for iRound in range(round_of_smooth):
            SM = iRound + 1

            logg.info("Smooth round:", SM)
            t = time.time()
            similarity_matrix = (
                beta * similarity_matrix + (1 - beta) * transpose_A * similarity_matrix
            )
            # similarity_matrix =beta*similarity_matrix+(1-beta)*similarity_matrix*adjacency_matrix
            # similarity_matrix_array.append(similarity_matrix)

            logg.hint("Time elapsed:", time.time() - t)

            t = time.time()
            sparsity_frac = (similarity_matrix > 0).sum() / (
                similarity_matrix.shape[0] * similarity_matrix.shape[1]
            )
            if sparsity_frac >= 0.1:
                # similarity_matrix_truncate=similarity_matrix
                # similarity_matrix_truncate_array.append(similarity_matrix_truncate)

                logg.hint(f"Orignal sparsity={sparsity_frac}, Thresholding")
                similarity_matrix = hf.matrix_row_or_column_thresholding(
                    similarity_matrix, truncation_threshold
                )
                sparsity_frac_2 = (similarity_matrix > 0).sum() / (
                    similarity_matrix.shape[0] * similarity_matrix.shape[1]
                )
                # similarity_matrix_truncate_array.append(similarity_matrix_truncate)

                logg.hint(f"Final sparsity={sparsity_frac_2}")

                logg.info(
                    f"similarity matrix truncated (Smooth round={SM}): ",
                    time.time() - t,
                )

            # logg.info("Save the matrix")
            # file_name=f'data/20200221_truncated_similarity_matrix_SM{round_of_smooth}_kNN{neighbor_N}_Truncate{str(truncation_threshold)[2:]}.npz'
            similarity_matrix = ssp.csr_matrix(similarity_matrix)

            ############## The new method
            # similarity_matrix=similarity_matrix.T.copy()
            ##############

            if save_subset:
                if SM % 5 == 0:  # save when SM=5,10,15,20,...

                    logg.hint("Save the matrix at every 5 rounds")
                    ssp.save_npz(file_name + f"_SM{SM}.npz", similarity_matrix)
            else:  # save all

                logg.hint("Save the matrix at every round")
                ssp.save_npz(file_name + f"_SM{SM}.npz", similarity_matrix)

    return similarity_matrix


def generate_initial_similarity(similarity_matrix, initial_index_0, initial_index_1):
    """
    Extract Smatrix at t1 from the full Smatrix

    Parameters
    ----------
    similarity_matrix: `np.array` or `sp.spmatrix`
        full Smatrix
    initial_index_0: `list`
        list of selected t1-cell id among all cells (t1+t2)
    initial_index_1: `list`
        list of selected t1-cell id among all cells (t1+t2)
        It can be the same as initial_index_0. In the case that they are different,
        initial_index_1 is a subset of cells that correspond to multi-time clones,
        while initial_index_0 may be all cells at t1.

    Returns
    -------
    initial Smatrix: `np.array`
    """

    t = time.time()
    initial_similarity = similarity_matrix[initial_index_0][:, initial_index_1]
    # initial_similarity=hf.sparse_column_multiply(initial_similarity,1/(resol+initial_similarity.sum(0)))
    if ssp.issparse(initial_similarity):
        initial_similarity = initial_similarity.A

    logg.hint("Time elapsed: ", time.time() - t)
    return initial_similarity


def generate_final_similarity(similarity_matrix, final_index_0, final_index_1):
    """
    Extract Smatrix at t2 from the full Smatrix

    Parameters
    ----------
    similarity_matrix: `np.array` or `sp.spmatrix`
        full Smatrix
    final_index_0: `list`
        list of selected t2-cell id among all cells (t1+t2)
    final_index_1: `list`
        list of selected t2-cell id among all cells (t1+t2)
        It can be the same as final_index_0. In the case that they are different,
        initial_index_0 is a subset of cells that correspond to multi-time clones,
        while initial_index_1 may be all cells at t2.

    Returns
    -------
    initial Smatrix: `np.array`
    """

    t = time.time()
    final_similarity = similarity_matrix.T[final_index_0][:, final_index_1]
    if ssp.issparse(final_similarity):
        final_similarity = final_similarity.A
    # final_similarity=hf.sparse_rowwise_multiply(final_similarity,1/(resol+final_similarity.sum(1)))

    logg.hint("Time elapsed: ", time.time() - t)
    return final_similarity


def select_time_points(
    adata_orig, time_point=["day_1", "day_2"], extend_Tmap_space=False
):
    """
    Select barcoded cells at given time points for Tmap inference.

    Select cells at given time points, and prepare the right data structure
    for running core cospar function to infer the Tmap.

    Parameters
    ----------
    adata_orig: original :class:`~anndata.AnnData` object
    time_point: `list` optional (default: ['day_1','day_2'])
        Require at least two time points, arranged in ascending order.
    extend_Tmap_space: `bool` optional (default: `False`)
        If true, the initial states for Tmap will include all states at initial time points,
        and the later states for Tmap will include all states at later time points.
        Otherwise, the initial and later state
        space of the Tmap will be restricted to cells with multi-time clonal information
        alone. The latter case speeds up the computation, which is recommended.

    Returns
    -------
    Subsampled :class:`~anndata.AnnData` object
    """

    # x_emb_orig=adata_orig.obsm['X_emb'][:,0]
    # y_emb_orig=adata_orig.obsm['X_emb'][:,1]
    time_info_orig = np.array(adata_orig.obs["time_info"])
    clone_annot_orig = adata_orig.obsm["X_clone"]
    if len(time_point) == 0:  # use all clonally labelled cell states
        time_point = np.sort(
            list(set(time_info_orig))
        )  # this automatic ordering may not work

    if len(time_point) < 2:
        logg.error("Must select more than 1 time point!")
    else:

        At = []
        for j, time_0 in enumerate(time_point):
            At.append(ssp.csr_matrix(clone_annot_orig[time_info_orig == time_0]))

        ### Day t - t+1
        Clonal_cell_ID_FOR_t = []
        for j in range(len(time_point) - 1):
            idx_t = np.array((At[j] * At[j + 1].T).sum(1) > 0).flatten()
            time_index_t = time_info_orig == time_point[j]
            temp = np.nonzero(time_index_t)[0][idx_t]
            Clonal_cell_ID_FOR_t.append(
                temp
            )  # this index is in the original space, without sampling etc

            logg.hint(
                f"Clonal cell fraction (day {time_point[j]}-{time_point[j+1]}):",
                len(temp) / np.sum(time_index_t),
            )

        ### Day t+1 - t
        Clonal_cell_ID_BACK_t = []
        for j in range(len(time_point) - 1):
            idx_t = np.array((At[j + 1] * At[j].T).sum(1) > 0).flatten()
            time_index_t = time_info_orig == time_point[j + 1]
            temp = np.nonzero(time_index_t)[0][idx_t]
            Clonal_cell_ID_BACK_t.append(
                temp
            )  # this index is in the original space, without sampling etc

            logg.hint(
                f"Clonal cell fraction (day {time_point[j+1]}-{time_point[j]}):",
                len(temp) / np.sum(time_index_t),
            )

        for j in range(len(time_point) - 1):
            logg.hint(
                f"Numer of cells that are clonally related -- day {time_point[j]}: {len(Clonal_cell_ID_FOR_t[j])}  and day {time_point[j+1]}: {len(Clonal_cell_ID_BACK_t[j])}"
            )

        proportion = np.ones(len(time_point))
        # flatten the list
        flatten_clonal_cell_ID_FOR = np.array(
            [sub_item for item in Clonal_cell_ID_FOR_t for sub_item in item]
        )
        flatten_clonal_cell_ID_BACK = np.array(
            [sub_item for item in Clonal_cell_ID_BACK_t for sub_item in item]
        )
        valid_clone_N_FOR = np.sum(
            clone_annot_orig[flatten_clonal_cell_ID_FOR].A.sum(0) > 0
        )
        valid_clone_N_BACK = np.sum(
            clone_annot_orig[flatten_clonal_cell_ID_BACK].A.sum(0) > 0
        )

        logg.info(f"Number of multi-time clones post selection: {valid_clone_N_FOR}")
        # logg.info("Valid clone number 'BACK' post selection",valid_clone_N_BACK)

        ###################### select initial and later cell states

        if extend_Tmap_space:
            old_Tmap_cell_id_t1 = []
            for t_temp in time_point[:-1]:
                old_Tmap_cell_id_t1 = old_Tmap_cell_id_t1 + list(
                    np.nonzero(time_info_orig == t_temp)[0]
                )
            old_Tmap_cell_id_t1 = np.array(old_Tmap_cell_id_t1)

            ########
            old_Tmap_cell_id_t2 = []
            for t_temp in time_point[1:]:
                old_Tmap_cell_id_t2 = old_Tmap_cell_id_t2 + list(
                    np.nonzero(time_info_orig == t_temp)[0]
                )
            old_Tmap_cell_id_t2 = np.array(old_Tmap_cell_id_t2)

        else:
            old_Tmap_cell_id_t1 = flatten_clonal_cell_ID_FOR
            old_Tmap_cell_id_t2 = flatten_clonal_cell_ID_BACK

        old_clonal_cell_id_t1 = flatten_clonal_cell_ID_FOR
        old_clonal_cell_id_t2 = flatten_clonal_cell_ID_BACK
        ########################

        sp_id = np.sort(
            list(set(list(old_Tmap_cell_id_t1) + list(old_Tmap_cell_id_t2)))
        )
        sp_idx = np.zeros(clone_annot_orig.shape[0], dtype=bool)
        sp_idx[sp_id] = True

        Tmap_cell_id_t1 = hf.converting_id_from_fullSpace_to_subSpace(
            old_Tmap_cell_id_t1, sp_id
        )[0]
        clonal_cell_id_t1 = hf.converting_id_from_fullSpace_to_subSpace(
            old_clonal_cell_id_t1, sp_id
        )[0]
        clonal_cell_id_t2 = hf.converting_id_from_fullSpace_to_subSpace(
            old_clonal_cell_id_t2, sp_id
        )[0]
        Tmap_cell_id_t2 = hf.converting_id_from_fullSpace_to_subSpace(
            old_Tmap_cell_id_t2, sp_id
        )[0]

        Clonal_cell_ID_FOR_t_new = []
        for temp_id_list in Clonal_cell_ID_FOR_t:
            convert_list = hf.converting_id_from_fullSpace_to_subSpace(
                temp_id_list, sp_id
            )[0]
            Clonal_cell_ID_FOR_t_new.append(convert_list)

        Clonal_cell_ID_BACK_t_new = []
        for temp_id_list in Clonal_cell_ID_BACK_t:
            convert_list = hf.converting_id_from_fullSpace_to_subSpace(
                temp_id_list, sp_id
            )[0]
            Clonal_cell_ID_BACK_t_new.append(convert_list)

        sp_id_0 = np.sort(list(old_clonal_cell_id_t1) + list(old_clonal_cell_id_t2))
        sp_idx_0 = np.zeros(clone_annot_orig.shape[0], dtype=bool)
        sp_idx_0[sp_id_0] = True

        barcode_id = np.nonzero(clone_annot_orig[sp_idx_0].A.sum(0).flatten() > 0)[0]
        # sp_id=np.nonzero(sp_idx)[0]
        clone_annot = clone_annot_orig[sp_idx][:, barcode_id]

        adata = adata_orig[sp_idx]
        adata.obsm["X_clone"] = clone_annot
        adata.uns["clonal_cell_id_t1"] = clonal_cell_id_t1
        adata.uns["clonal_cell_id_t2"] = clonal_cell_id_t2
        adata.uns["Tmap_cell_id_t1"] = Tmap_cell_id_t1
        adata.uns["Tmap_cell_id_t2"] = Tmap_cell_id_t2
        adata.uns["multiTime_cell_id_t1"] = Clonal_cell_ID_FOR_t_new
        adata.uns["multiTime_cell_id_t2"] = Clonal_cell_ID_BACK_t_new
        adata.uns["proportion"] = np.ones(len(time_point) - 1)
        adata.uns["sp_idx"] = sp_idx

        data_des_orig = adata_orig.uns["data_des"][0]
        data_des_0 = adata_orig.uns["data_des"][-1]
        time_label = "t"
        for x in time_point:
            time_label = time_label + f"*{x}"

        data_des = (
            data_des_0
            + f"_MultiTimeClone_FullSpace{int(extend_Tmap_space)}_{time_label}"
        )
        adata.uns["data_des"] = [data_des_orig, data_des]

        if logg._settings_verbosity_greater_or_equal_than(3):
            N_cell, N_clone = clone_annot.shape
            logg.info(f"Cell number={N_cell}, Clone number={N_clone}")
            x_emb = adata.obsm["X_emb"][:, 0]
            y_emb = adata.obsm["X_emb"][:, 1]
            pl.customized_embedding(x_emb, y_emb, -x_emb)

        logg.hint(f"clonal_cell_id_t1: {len(clonal_cell_id_t1)}")
        logg.hint(f"Tmap_cell_id_t1: {len(Tmap_cell_id_t1)}")
        return adata



import os

import numpy as np
import pandas as pd
import scipy.sparse as ssp
import scipy.stats as stats
import seaborn as sns
import statsmodels.sandbox.stats.multicomp
from ete3 import Tree
from matplotlib import pyplot as plt
from numpy.lib.twodim_base import tril_indices
from scipy.cluster import hierarchy

# from plotnine import *
from sklearn.manifold import SpectralEmbedding

from cospar import tool as tl
from cospar.plotting import _utils as pl_util

from .. import help_functions as hf
from .. import logging as logg
from .. import settings


def barcode_heatmap(
    adata,
    selected_times=None,
    selected_fates=None,
    color_bar=True,
    rename_fates=None,
    normalize=False,
    binarize=False,
    log_transform=False,
    fig_width=4,
    fig_height=6,
    figure_index="",
    plot=True,
    pseudocount=10 ** (-10),
    order_map_x=False,
    order_map_y=False,
    fate_normalize_source="X_clone",
    select_clones_with_fates: list = None,
    select_clones_without_fates: list = None,
    select_clones_mode: str = "or",
    **kwargs,
):
    """
    Plot barcode heatmap among different fate clusters.

    We clonal measurement at selected time points and show the
    corresponding heatmap among selected fate clusters.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    selected_times: `list`, optional (default: None)
        Time points to select the cell states.
    selected_fates: `list`, optional (default: all)
        List of fate clusters to use. If set to be [], use all.
    color_bar: `bool`, optional (default: True)
        Plot color bar.
    rename_fates: `list`, optional (default: None)
        Provide new names in substitution of names in selected_fates.
        For this to be effective, the new name list needs to have names
        in exact correspondence to those in the old list.
    normalize:
        To perform cluster-wise then clone-wise normalization per time point
    binarize: `bool`
        Binarize the coarse-grained barcode count matrix, just for the purpose of plotting.
    log_transform: `bool`, optional (default: False)
        If true, perform a log transform. This is needed when the data
        matrix has entries varying by several order of magnitude.
    fig_width: `float`, optional (default: 4)
        Figure width.
    fig_height: `float`, optional (default: 6)
        Figure height.
    plot: `bool`
        True: plot the result. False, suppress the plot.
    pseudocount: `float`
        Pseudocount for the heatmap (needed for ordering the map)
    order_map_x: `bool`
        Whether to re-order the x coordinate of the matrix or not
    order_map_y: `bool`
        Whether to re-order the y coordinate of the matrix or not
    fate_normalize_source:
        Source for cluster-wise normalization: {'X_clone','state_info'}. 'X_clone': directly row-normalize coarse_X_clone; 'state_info': compute each cluster size directly, and then normalize coarse_X_clone. The latter method is useful if we have single-cell resolution for each fate.
    select_clones_with_fates: list = None,
        Select clones that labels fates from this list.
    select_clones_without_fates: list = None,
        Exclude clones that labels fates from this list.
    select_clones_mode: str = {'or','and'}
        Logic rule for selection.

    Returns:
    --------
    The coarse-grained X_clone matrix and the selected clusters are returned at
    adata.uns['barcode_heatmap']. The coarse-grained X_clone keeps all clones and maintains their ordering.
    """

    data_des = adata.uns["data_des"][-1]
    data_des = f"{data_des}_clonal"
    figure_path = settings.figure_path

    coarse_X_clone, mega_cluster_list = tl.coarse_grain_clone_over_cell_clusters(
        adata,
        selected_times=selected_times,
        selected_fates=selected_fates,
        normalize=normalize,
        fate_normalize_source=fate_normalize_source,
        select_clones_with_fates=select_clones_with_fates,
        select_clones_without_fates=select_clones_without_fates,
        select_clones_mode=select_clones_mode,
        **kwargs,
    )

    if rename_fates is None:
        rename_fates = mega_cluster_list

    if len(rename_fates) != len(mega_cluster_list):
        logg.warn(
            "rename_fates does not have the same length as selected_fates, thus not used."
        )
        rename_fates = mega_cluster_list

    if "x_ticks" not in kwargs.keys():
        kwargs["x_ticks"] = rename_fates

    coarse_X_clone_new = pl_util.custom_hierachical_ordering(
        np.arange(coarse_X_clone.shape[0]), coarse_X_clone
    )
    adata.uns["barcode_heatmap"] = {
        "coarse_X_clone": coarse_X_clone,
        "fate_names": rename_fates,
    }
    logg.info("Data saved at adata.uns['barcode_heatmap']")
    if plot:
        if binarize:
            final_matrix = coarse_X_clone_new > 0
            color_bar_label = "Binarized barcode count"
        else:
            final_matrix = coarse_X_clone_new
            color_bar_label = "Barcode count"

        if normalize:
            color_bar_label += " (normalized)"

        clone_idx = final_matrix.sum(0) > 0
        ax = pl_util.heatmap(
            final_matrix[:, clone_idx].T + pseudocount,
            order_map_x=order_map_x,
            order_map_y=order_map_y,
            color_bar_label=color_bar_label,
            log_transform=log_transform,
            fig_width=fig_width,
            fig_height=fig_height,
            color_bar=color_bar,
            **kwargs,
        )
        plt.title(f"{np.sum(clone_idx)} clones")

        plt.tight_layout()
        if figure_index != "":
            figure_index == f"_{figure_index}"
        plt.savefig(
            os.path.join(
                figure_path,
                f"{data_des}_barcode_heatmap{figure_index}.{settings.file_format_figs}",
            )
        )
        return ax


def clonal_fates_across_time(adata, selected_times, **kwargs):
    """
    Scatter plot for clonal fate number across time point

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    selected_times: `list`, optional (default: None)
        Time points to select the cell states.

    Returns
    -------
    Results updated at adata.uns["clonal_fates_across_time"]
    """
    if len(selected_times) != 2:
        raise ValueError("selected_times must be a list with two values")
    barcode_heatmap(
        adata,
        selected_times=selected_times[0],
        color_bar=True,
        log_transform=False,
        plot=False,
    )
    clonal_fates_t1 = (adata.uns["barcode_heatmap"]["coarse_X_clone"] > 0).sum(0)
    barcode_heatmap(
        adata,
        selected_times=selected_times[1],
        color_bar=True,
        log_transform=False,
        plot=False,
    )
    clonal_fates_t2 = (adata.uns["barcode_heatmap"]["coarse_X_clone"] > 0).sum(0)

    pl_util.jitter(clonal_fates_t1, clonal_fates_t2, **kwargs)
    plt.xlabel(f"Number of fates per clone (t={selected_times[0]})")
    plt.ylabel(f"Number of fates per clone (t={selected_times[1]})")
    data_des = adata.uns["data_des"][0]
    plt.savefig(
        os.path.join(
            settings.figure_path,
            f"{data_des}_barcode_coupling_across_time.{settings.file_format_figs}",
        )
    )
    adata.uns["clonal_fates_across_time"] = {
        "clonal_fates_t1": clonal_fates_t1,
        "clonal_fates_t2": clonal_fates_t2,
    }
    logg.info("Data saved at adata.uns['clonal_fates_across_time']")


def clones_on_manifold(
    adata,
    selected_clone_list=[0],
    color_list=["red", "blue", "purple", "green", "cyan", "black"],
    selected_times=None,
    title=None,
    clone_markersize=12,
    clone_markeredgewidth=1,
    markeredgecolor="black",
    **kwargs,
):
    """
    Plot clones on top of state embedding.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    selected_clone_list: `list`
        List of selected clone ID's.
    color_list: `list`, optional (default: ['red','blue','purple','green','cyan','black'])
        The list of color that defines color at respective time points.
    selected_times: `list`, optional (default: all)
        Select time points to show corresponding states. If set to be [], use all states.
    title: None
        Default, show the clone id as panel title. Otherwise, show the text string from input 'title'
    clone_markersize: `int`, optional (default: 12)
        Clone marker size
    clone_markeredgewidth: `int`, optional (default: 1)
        Edige size for clone marker
    """

    fig_width = settings.fig_width
    fig_height = settings.fig_height
    point_size = settings.fig_point_size
    x_emb = adata.obsm["X_emb"][:, 0]
    y_emb = adata.obsm["X_emb"][:, 1]
    data_des = adata.uns["data_des"][-1]
    # data_path=settings.data_path
    figure_path = settings.figure_path
    X_clone = adata.obsm["X_clone"]
    time_info = np.array(adata.obs["time_info"])

    # use only valid time points
    sp_idx = hf.selecting_cells_by_time_points(time_info, selected_times)
    selected_times = np.sort(list(set(time_info[sp_idx])))

    selected_clone_list = np.array(selected_clone_list)
    full_id_list = np.arange(X_clone.shape[1])
    valid_idx = np.in1d(full_id_list, selected_clone_list)
    if np.sum(valid_idx) < len(selected_clone_list):
        logg.error(
            f"Valid id range is (0,{X_clone.shape[1]-1}). Please use a smaller ID!"
        )
        selected_clone_list = full_id_list[valid_idx]

    if len(selected_clone_list) == 0:
        logg.error("No valid states selected.")
    else:
        # using all data
        for my_id in selected_clone_list:
            fig = plt.figure(figsize=(fig_width, fig_height))
            ax = plt.subplot(1, 1, 1)
            idx_t = np.zeros(len(time_info), dtype=bool)
            for j, xx in enumerate(selected_times):
                idx_t0 = time_info == selected_times[j]
                idx_t = idx_t0 | idx_t

            pl_util.customized_embedding(
                x_emb[idx_t],
                y_emb[idx_t],
                np.zeros(len(y_emb[idx_t])),
                ax=ax,
                point_size=point_size,
            )
            for j, xx in enumerate(selected_times):
                idx_t = time_info == selected_times[j]
                idx_clone = X_clone[:, my_id].A.flatten() > 0
                idx = idx_t & idx_clone
                ax.plot(
                    x_emb[idx],
                    y_emb[idx],
                    ".",
                    color=color_list[j % len(color_list)],
                    markersize=clone_markersize,
                    markeredgecolor=markeredgecolor,
                    markeredgewidth=clone_markeredgewidth,
                    **kwargs,
                )

                if title is None:
                    ax.set_title(f"ID: {my_id}")
                else:
                    ax.set_title(title)

            fig.savefig(
                os.path.join(
                    figure_path,
                    f"{data_des}_different_clones_{my_id}.{settings.file_format_figs}",
                )
            )
    return ax


def clonal_fate_bias(adata, show_histogram=True, FDR=0.05):
    """
    Plot clonal fate bias towards a cluster.

    The results should be pre-computed from :func:`cospar.tl.clonal_fate_bias`

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    show_histogram: `bool`, optional (default: True)
        If true, show the distribution of inferred fate probability.
    FDR: `float`, optional (default: 0.05)
        False-discovery rate after the Benjamini-Hochberg correction.
    """

    if "clonal_fate_bias" not in adata.uns.keys():
        raise ValueError(
            "clonal_fate_bias has not been computed. Run cs.tl.clonal_fate_bias first"
        )
    else:
        df = adata.uns["clonal_fate_bias"]
        fate_bias = df["Fate_bias"]
        target_fraction_array = df["clonal_fraction_in_target_fate"]

    fig_width = settings.fig_width
    fig_height = settings.fig_height
    data_des = adata.uns["data_des"][-1]
    figure_path = settings.figure_path
    FDR_threshold = -np.log10(FDR)

    fig = plt.figure(figsize=(fig_width, fig_height))
    ax = plt.subplot(1, 1, 1)
    ax.plot(np.arange(len(fate_bias)), fate_bias, ".", color="blue", markersize=5)
    ax.plot(
        np.arange(len(fate_bias)),
        np.zeros(len(fate_bias)) + FDR_threshold,
        "-.",
        color="grey",
        markersize=5,
        label=f"FDR={FDR}",
    )

    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    # ax.set_ylabel('Fate bias ($-\\log_{10}P_{value}$)')
    ax.set_ylabel("Clonal fate bias")
    ax.set_xlabel("Clonal index")
    ax.legend()
    fig.tight_layout()
    fig.savefig(
        os.path.join(
            figure_path, f"{data_des}_clonal_fate_bias.{settings.file_format_figs}"
        )
    )

    if show_histogram:
        fig = plt.figure(figsize=(fig_width, fig_height))
        ax = plt.subplot(1, 1, 1)
        ax.hist(target_fraction_array, color="#2ca02c", density=True)
        ax.set_xlim([0, 1])
        ax.set_xlabel("Clonal fraction in selected fates")
        ax.set_ylabel("Density")
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.set_title(f"Average: {np.mean(target_fraction_array):.2f}")
        fig.tight_layout()
        fig.savefig(
            os.path.join(
                figure_path,
                f"{data_des}_observed_clonal_fraction.{settings.file_format_figs}",
            )
        )


def clonal_reports(adata, selected_times=None, **kwargs):
    """
    Report the statistics of the clonal data.

    It includes the statistics for clone size , and the barcode number per cell.
    """

    time_info = np.array(adata.obs["time_info"])
    sp_idx = hf.selecting_cells_by_time_points(time_info, selected_times)
    adata_1 = adata[sp_idx]
    persistent_clone_ids = tl.identify_persistent_clones(adata_1)
    X_clone = adata_1.obsm["X_clone"]
    total_clone_N = X_clone.shape[1]
    print(
        f"  Clones observed across selected times: {len(persistent_clone_ids)} (out of {total_clone_N} clones)"
    )

    for x in set(adata_1.obs["time_info"]):
        print(f"---------t={x}---------")
        adata_sp = adata_1[adata_1.obs["time_info"] == x]
        X_clone = adata_sp.obsm["X_clone"]
        clone_size = X_clone.sum(0).A.flatten()
        clonal_bc_number = X_clone.sum(1).A.flatten()
        clonal_cells_N = np.sum(clonal_bc_number > 0)
        total_N = X_clone.shape[0]
        total_clone_N = X_clone.shape[1]
        useful_clone_N = np.sum(clone_size > 0)
        print(f"    Cells with barcode: {clonal_cells_N} (out of {total_N} cells)")
        print(
            f"    Barcodes with cells: {useful_clone_N} (out of {total_clone_N} clones)"
        )

        fig, axs = plt.subplots(1, 2, figsize=(8, 3.5))
        ax = sns.histplot(clone_size[clone_size > 0], ax=axs[0], **kwargs)
        ax.set_xlabel("Clone size")
        ax.set_ylabel("Count")

        ax = sns.histplot(clonal_bc_number[clonal_bc_number > 0], ax=axs[1], **kwargs)
        ax.set_xlabel("Clonal barcode number per cell")
        ax.set_ylabel("Count")
        fig.suptitle(f"Time={x}")



#Hematopoiesis_distance_matrix用于Hematopoiesis的演示，Reprogramming_distance_matrix用于Reprogramming的演示，distance_matrix()是自定义样本的计算模式

import sys
import numpy as np
import matplotlib.pyplot as plt
import os
import time

import numpy as np
import pandas as pd
import scanpy as sc
import scipy.sparse as ssp

from cospar.tmap import _tmap_core as tmap_core
from cospar.tmap import _utils as tmap_util

from cospar import help_functions as hf
from cospar import logging as logg
from cospar import settings
from cospar import tool as tl
from types import MappingProxyType
from typing import (
    Union,
    Optional,
    Any,
    Mapping,
    Callable,
    NamedTuple,
    Generator,
    Tuple,
    Literal,
)
import warnings

import numpy as np
import scipy
from anndata import AnnData
from scipy.sparse import issparse, coo_matrix, csr_matrix
from sklearn.utils import check_random_state

def Hematopoiesis_distance_matrix(
    adata,
    round_of_smooth=1,
    neighbor_N=30,
    beta=0.1,
    truncation_threshold=0.001,
    save_subset=True,
    use_existing_KNN_graph=False,
    compute_new_Smatrix=True,
    use_full_Smatrix = True,
):
    """
    Generate similarity matrix (Smatrix) through graph diffusion

    It generates the similarity matrix via iterative graph diffusion.
    Similarity matrix from each round of diffusion will be saved, after truncation
    to promote sparsity and save space. If save_subset is activated, only save
    Smatrix for smooth rounds at the multiples of 5 (like 5,10,15,...). If a Smatrix is pre-computed,
    it will be loaded directly if compute_new_Smatrix=Flase.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    file_name: str
        Filename to load pre-computed similarity matrix or save the newly
        computed similarity matrix.
    round_of_smooth: `int`, optional (default: 10)
        The rounds of graph diffusion.
    neighbor_N: `int`, optional (default: 20)
        Neighber number for constructing the KNN graph, using the UMAP method.
    beta: `float`, option (default: 0.1)
        Probability to stay at the origin in a unit diffusion step, in the range [0,1]
    truncation_threshold: `float`, optional (default: 0.001)
        At each iteration, truncate the similarity matrix using
        truncation_threshold. This promotes the sparsity of the matrix,
        thus the speed of computation. We set the truncation threshold to be small,
        to guarantee accracy.
    save_subset: `bool`, optional (default: True)
        If true, save only Smatrix at smooth round [5,10,15,...];
        Otherwise, save Smatrix at each round.
    use_existing_KNN_graph: `bool`, optional (default: False)
        If true and adata.obsp['connectivities'], use the existing knn graph to build
        the similarity matrix, regardless of neighbor_N.
    compute_new_Smatrix: `bool`, optional (default: False)
        If true, compute a new Smatrix, even if there is pre-computed Smatrix with the
        same parameterization.

    Returns
    -------
        similarity_matrix: `sp.spmatrix`
    """




            
    if (1):  # compute now
        from cospar import logging as logg
        logg.hint(f"Compute similarity matrix: computing new; beta={beta}")

        # add a step to compute PCA in case this is not computed

        if (not use_existing_KNN_graph) or ("connectivities" not in adata.obsp.keys()):
            # here, we assume that adata already has pre-computed PCA
            n_neighbors=neighbor_N
            sc.pp.neighbors(adata, n_neighbors=neighbor_N)###################################################################
        
        else:
            logg.hint(
                "Use existing KNN graph at adata.obsp['connectivities'] for generating the smooth matrix"
            )
        adjacency_matrix = adata.obsp["connectivities"]

        ############## The new method
        adjacency_matrix = (adjacency_matrix + adjacency_matrix.T) / 2
        ##############

        adjacency_matrix = hf.sparse_rowwise_multiply(
            adjacency_matrix, 1 / adjacency_matrix.sum(1).A.squeeze()
        )
        nrow = adata.shape[0]
        similarity_matrix = ssp.lil_matrix((nrow, nrow))
        similarity_matrix.setdiag(np.ones(nrow))
        transpose_A = adjacency_matrix.T

        if round_of_smooth == 0:
            SM = 0
            similarity_matrix = ssp.csr_matrix(similarity_matrix)
            ##ssp.save_npz(file_name + f"_SM{SM}.npz", similarity_matrix)

        for iRound in range(round_of_smooth):
            SM = iRound + 1

            logg.info("Smooth round:", SM)
            t = time.time()
            similarity_matrix = (
                beta * similarity_matrix + (1 - beta) * transpose_A * similarity_matrix
            )
            # similarity_matrix =beta*similarity_matrix+(1-beta)*similarity_matrix*adjacency_matrix
            # similarity_matrix_array.append(similarity_matrix)

            logg.hint("Time elapsed:", time.time() - t)

            t = time.time()
            sparsity_frac = (similarity_matrix > 0).sum() / (
                similarity_matrix.shape[0] * similarity_matrix.shape[1]
            )
            if sparsity_frac >= 0.1:
                # similarity_matrix_truncate=similarity_matrix
                # similarity_matrix_truncate_array.append(similarity_matrix_truncate)

                logg.hint(f"Orignal sparsity={sparsity_frac}, Thresholding")
                similarity_matrix = hf.matrix_row_or_column_thresholding(
                    similarity_matrix, truncation_threshold
                )
                sparsity_frac_2 = (similarity_matrix > 0).sum() / (
                    similarity_matrix.shape[0] * similarity_matrix.shape[1]
                )
                # similarity_matrix_truncate_array.append(similarity_matrix_truncate)

                logg.hint(f"Final sparsity={sparsity_frac_2}")

                logg.info(
                    f"similarity matrix truncated (Smooth round={SM}): ",
                    time.time() - t,
                )

            # logg.info("Save the matrix")
            # file_name=f'data/20200221_truncated_similarity_matrix_SM{round_of_smooth}_kNN{neighbor_N}_Truncate{str(truncation_threshold)[2:]}.npz'
            similarity_matrix = ssp.csr_matrix(similarity_matrix)

            ############## The new method
            # similarity_matrix=similarity_matrix.T.copy()
            ##############

            if save_subset:
                if SM % 5 == 0:  # save when SM=5,10,15,20,...

                    logg.hint("Save the matrix at every 5 rounds")
                    ##ssp.save_npz(file_name + f"_SM{SM}.npz", similarity_matrix)
            else:  # save all

                logg.hint("Save the matrix at every round")
                ##ssp.save_npz(file_name + f"_SM{SM}.npz", similarity_matrix)
    distance_matrix = similarity_matrix
    return distance_matrix



def Reprogramming_distance_matrix(
    adata,
    round_of_smooth=15,
    neighbor_N=20,
    beta=0.1,
    truncation_threshold=0.001,
    save_subset=True,
    use_existing_KNN_graph=False,
    compute_new_Smatrix=True,
    use_full_Smatrix = True,
):
    """
    Generate similarity matrix (Smatrix) through graph diffusion

    It generates the similarity matrix via iterative graph diffusion.
    Similarity matrix from each round of diffusion will be saved, after truncation
    to promote sparsity and save space. If save_subset is activated, only save
    Smatrix for smooth rounds at the multiples of 5 (like 5,10,15,...). If a Smatrix is pre-computed,
    it will be loaded directly if compute_new_Smatrix=Flase.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    file_name: str
        Filename to load pre-computed similarity matrix or save the newly
        computed similarity matrix.
    round_of_smooth: `int`, optional (default: 10)
        The rounds of graph diffusion.
    neighbor_N: `int`, optional (default: 20)
        Neighber number for constructing the KNN graph, using the UMAP method.
    beta: `float`, option (default: 0.1)
        Probability to stay at the origin in a unit diffusion step, in the range [0,1]
    truncation_threshold: `float`, optional (default: 0.001)
        At each iteration, truncate the similarity matrix using
        truncation_threshold. This promotes the sparsity of the matrix,
        thus the speed of computation. We set the truncation threshold to be small,
        to guarantee accracy.
    save_subset: `bool`, optional (default: True)
        If true, save only Smatrix at smooth round [5,10,15,...];
        Otherwise, save Smatrix at each round.
    use_existing_KNN_graph: `bool`, optional (default: False)
        If true and adata.obsp['connectivities'], use the existing knn graph to build
        the similarity matrix, regardless of neighbor_N.
    compute_new_Smatrix: `bool`, optional (default: False)
        If true, compute a new Smatrix, even if there is pre-computed Smatrix with the
        same parameterization.

    Returns
    -------
        similarity_matrix: `sp.spmatrix`
    """




            
    if (1):  # compute now
        from cospar import logging as logg
        logg.hint(f"Compute similarity matrix: computing new; beta={beta}")

        # add a step to compute PCA in case this is not computed

        if (not use_existing_KNN_graph) or ("connectivities" not in adata.obsp.keys()):
            # here, we assume that adata already has pre-computed PCA
            n_neighbors=neighbor_N
            sc.pp.neighbors(adata, n_neighbors=neighbor_N)###################################################################
        
        else:
            logg.hint(
                "Use existing KNN graph at adata.obsp['connectivities'] for generating the smooth matrix"
            )
        adjacency_matrix = adata.obsp["connectivities"]

        ############## The new method
        adjacency_matrix = (adjacency_matrix + adjacency_matrix.T) / 2
        ##############

        adjacency_matrix = hf.sparse_rowwise_multiply(
            adjacency_matrix, 1 / adjacency_matrix.sum(1).A.squeeze()
        )
        nrow = adata.shape[0]
        similarity_matrix = ssp.lil_matrix((nrow, nrow))
        similarity_matrix.setdiag(np.ones(nrow))
        transpose_A = adjacency_matrix.T

        if round_of_smooth == 0:
            SM = 0
            similarity_matrix = ssp.csr_matrix(similarity_matrix)
            ##ssp.save_npz(file_name + f"_SM{SM}.npz", similarity_matrix)

        for iRound in range(round_of_smooth):
            SM = iRound + 1

            logg.info("Smooth round:", SM)
            t = time.time()
            similarity_matrix = (
                beta * similarity_matrix + (1 - beta) * transpose_A * similarity_matrix
            )
            # similarity_matrix =beta*similarity_matrix+(1-beta)*similarity_matrix*adjacency_matrix
            # similarity_matrix_array.append(similarity_matrix)

            logg.hint("Time elapsed:", time.time() - t)

            t = time.time()
            sparsity_frac = (similarity_matrix > 0).sum() / (
                similarity_matrix.shape[0] * similarity_matrix.shape[1]
            )
            if sparsity_frac >= 0.1:
                # similarity_matrix_truncate=similarity_matrix
                # similarity_matrix_truncate_array.append(similarity_matrix_truncate)

                logg.hint(f"Orignal sparsity={sparsity_frac}, Thresholding")
                similarity_matrix = hf.matrix_row_or_column_thresholding(
                    similarity_matrix, truncation_threshold
                )
                sparsity_frac_2 = (similarity_matrix > 0).sum() / (
                    similarity_matrix.shape[0] * similarity_matrix.shape[1]
                )
                # similarity_matrix_truncate_array.append(similarity_matrix_truncate)

                logg.hint(f"Final sparsity={sparsity_frac_2}")

                logg.info(
                    f"similarity matrix truncated (Smooth round={SM}): ",
                    time.time() - t,
                )

            # logg.info("Save the matrix")
            # file_name=f'data/20200221_truncated_similarity_matrix_SM{round_of_smooth}_kNN{neighbor_N}_Truncate{str(truncation_threshold)[2:]}.npz'
            similarity_matrix = ssp.csr_matrix(similarity_matrix)

            ############## The new method
            # similarity_matrix=similarity_matrix.T.copy()
            ##############

            if save_subset:
                if SM % 5 == 0:  # save when SM=5,10,15,20,...

                    logg.hint("Save the matrix at every 5 rounds")
                    ##ssp.save_npz(file_name + f"_SM{SM}.npz", similarity_matrix)
            else:  # save all

                logg.hint("Save the matrix at every round")
                ##ssp.save_npz(file_name + f"_SM{SM}.npz", similarity_matrix)
    distance_matrix = similarity_matrix
    return distance_matrix

#----------------------------------------------------------------------------------------------------------------
def distance_matrix(
    adata,
    round_of_smooth,
    neighbor_N,
    beta,
    truncation_threshold,
    save_subset,
    use_existing_KNN_graph=False,
    compute_new_Smatrix=True,
    use_full_Smatrix = True,
):
    """
    Generate similarity matrix (Smatrix) through graph diffusion

    It generates the similarity matrix via iterative graph diffusion.
    Similarity matrix from each round of diffusion will be saved, after truncation
    to promote sparsity and save space. If save_subset is activated, only save
    Smatrix for smooth rounds at the multiples of 5 (like 5,10,15,...). If a Smatrix is pre-computed,
    it will be loaded directly if compute_new_Smatrix=Flase.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    file_name: str
        Filename to load pre-computed similarity matrix or save the newly
        computed similarity matrix.
    round_of_smooth: `int`, optional (default: 10)
        The rounds of graph diffusion.
    neighbor_N: `int`, optional (default: 20)
        Neighber number for constructing the KNN graph, using the UMAP method.
    beta: `float`, option (default: 0.1)
        Probability to stay at the origin in a unit diffusion step, in the range [0,1]
    truncation_threshold: `float`, optional (default: 0.001)
        At each iteration, truncate the similarity matrix using
        truncation_threshold. This promotes the sparsity of the matrix,
        thus the speed of computation. We set the truncation threshold to be small,
        to guarantee accracy.
    save_subset: `bool`, optional (default: True)
        If true, save only Smatrix at smooth round [5,10,15,...];
        Otherwise, save Smatrix at each round.
    use_existing_KNN_graph: `bool`, optional (default: False)
        If true and adata.obsp['connectivities'], use the existing knn graph to build
        the similarity matrix, regardless of neighbor_N.
    compute_new_Smatrix: `bool`, optional (default: False)
        If true, compute a new Smatrix, even if there is pre-computed Smatrix with the
        same parameterization.

    Returns
    -------
        similarity_matrix: `sp.spmatrix`
    """




            
    if (1):  # compute now
        from cospar import logging as logg
        logg.hint(f"Compute similarity matrix: computing new; beta={beta}")

        # add a step to compute PCA in case this is not computed

        if (not use_existing_KNN_graph) or ("connectivities" not in adata.obsp.keys()):
            # here, we assume that adata already has pre-computed PCA
            n_neighbors=neighbor_N
            sc.pp.neighbors(adata, n_neighbors=neighbor_N)###################################################################
        
        else:
            logg.hint(
                "Use existing KNN graph at adata.obsp['connectivities'] for generating the smooth matrix"
            )
        adjacency_matrix = adata.obsp["connectivities"]

        ############## The new method
        adjacency_matrix = (adjacency_matrix + adjacency_matrix.T) / 2
        ##############

        adjacency_matrix = hf.sparse_rowwise_multiply(
            adjacency_matrix, 1 / adjacency_matrix.sum(1).A.squeeze()
        )
        nrow = adata.shape[0]
        similarity_matrix = ssp.lil_matrix((nrow, nrow))
        similarity_matrix.setdiag(np.ones(nrow))
        transpose_A = adjacency_matrix.T

        if round_of_smooth == 0:
            SM = 0
            similarity_matrix = ssp.csr_matrix(similarity_matrix)
            ##ssp.save_npz(file_name + f"_SM{SM}.npz", similarity_matrix)

        for iRound in range(round_of_smooth):
            SM = iRound + 1

            logg.info("Smooth round:", SM)
            t = time.time()
            similarity_matrix = (
                beta * similarity_matrix + (1 - beta) * transpose_A * similarity_matrix
            )
            # similarity_matrix =beta*similarity_matrix+(1-beta)*similarity_matrix*adjacency_matrix
            # similarity_matrix_array.append(similarity_matrix)

            logg.hint("Time elapsed:", time.time() - t)

            t = time.time()
            sparsity_frac = (similarity_matrix > 0).sum() / (
                similarity_matrix.shape[0] * similarity_matrix.shape[1]
            )
            if sparsity_frac >= 0.1:
                # similarity_matrix_truncate=similarity_matrix
                # similarity_matrix_truncate_array.append(similarity_matrix_truncate)

                logg.hint(f"Orignal sparsity={sparsity_frac}, Thresholding")
                similarity_matrix = hf.matrix_row_or_column_thresholding(
                    similarity_matrix, truncation_threshold
                )
                sparsity_frac_2 = (similarity_matrix > 0).sum() / (
                    similarity_matrix.shape[0] * similarity_matrix.shape[1]
                )
                # similarity_matrix_truncate_array.append(similarity_matrix_truncate)

                logg.hint(f"Final sparsity={sparsity_frac_2}")

                logg.info(
                    f"similarity matrix truncated (Smooth round={SM}): ",
                    time.time() - t,
                )

            # logg.info("Save the matrix")
            # file_name=f'data/20200221_truncated_similarity_matrix_SM{round_of_smooth}_kNN{neighbor_N}_Truncate{str(truncation_threshold)[2:]}.npz'
            similarity_matrix = ssp.csr_matrix(similarity_matrix)

            ############## The new method
            # similarity_matrix=similarity_matrix.T.copy()
            ##############

            if save_subset:
                if SM % 5 == 0:  # save when SM=5,10,15,20,...

                    logg.hint("Save the matrix at every 5 rounds")
                    ##ssp.save_npz(file_name + f"_SM{SM}.npz", similarity_matrix)
            else:  # save all

                logg.hint("Save the matrix at every round")
                ##ssp.save_npz(file_name + f"_SM{SM}.npz", similarity_matrix)
    distance_matrix = similarity_matrix
    return distance_matrix